#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.reffram contains functions for operations
related to GNSS-products


it can be imported directly with:
from geodezyx import reffram

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import itertools

#### Import the logger
import logging
import os
import re

import matplotlib
import matplotlib.pyplot as plt
import natsort
import numpy as np
import pandas as pd

#### geodeZYX modules
from geodezyx import conv
from geodezyx import files_rw
from geodezyx import stats
from geodezyx import utils
from geodezyx.reffram import kepler_gzyx

### disabled and imported directly in the needed fct
## import geodezyx.reffram.sofa18 as sofa
log = logging.getLogger("geodezyx")

##########  END IMPORT  ##########


def compar_orbit(
    data_inp_1,
    data_inp_2,
    step_data=900,
    sats_used_list=["G"],
    name1="",
    name2="",
    use_name_1_2_for_table_name=False,
    rtn_output=True,
    convert_ecef_eci=True,
    clean_null_values=True,
    conv_coef=10**3,
    return_sat_null=False,
):
    """
    Compares 2 GNSS orbits files (SP3), and gives a summary plot and a
    statistics table

    Parameters
    ----------
    data_inp_1, data_inp_2 : str or Pandas DataFrame
        contains the orbits or path (string) to the sp3

    step_data : int
        per default data sampling

    sats_used_list : list of str
        used constellation or satellite : G E r c ... E01 , G02 ...
        Individuals satellites are prioritary on whole constellations
        e.g. ['G',"E04"]

    rtn_output : bool
        select output, Radial Transverse Normal or XYZ

    convert_ecef_eci : bool
        convert sp3 ECEF => ECI (Terrestrial => Celestrial)
        must be True in operational to avoid artifacts.

    name1, name2 : str (optionals)
        optional custom names for the 2 orbits

    use_name_1_2_for_table_name : bool
        False : use name 1 and 2 for table name, use datafile instead

    clean_null_values : bool or str
        if True or "all" remove sat position in all X,Y,Z values
        are null (0.000000)
        if "any", remove sat position if X or Y or Z is null
        if False, keep everything

    conv_coef : int
        conversion coefficient, km to m 10**3, km to mm 10**6

    Returns
    -------
    diff_sat_all : Pandas DataFrame
    contains differences b/w data_inp_1 & data_inp_2
    in Radial Transverse Normal OR XYZ frame

        Attributes of diff_sat_all :
            diff_sat_all.name : title of the table

    Note
    ----
    clean_null_values if useful (and necessary) only if
    convert_ECEF_ECI = False
    if convert_ECEF_ECI = True, the cleaning will be done by
    a side effect trick : the convertion ECEF => ECI will generate NaN
    for a zero-valued position
    But, nevertheless, activating  clean_null_values = True is better
    This Note is in fact usefull if you want to see bad positions on a plot
    => Then convert_ECEF_ECI = False and clean_null_values = False

    References
    ----------
    "Coordinate Systems", ASEN 3200 1/24/06 George h. Born

    """

    # selection of both used Constellations AND satellites
    sys_used_list = []
    prn_used_list = []
    for e in sats_used_list:
        if len(e) == 1:  ## it is a constellation
            sys_used_list.append(e)
        elif len(e) == 3:  ## it is a satellite
            prn_used_list.append(e)
            if not e[0] in sys_used_list:
                sys_used_list.append(e[0])

    # Read the files or DataFrames
    # metadata attributes are not copied
    # Thus, manual copy ...
    # (Dirty way, should be impoved without so many lines ...)
    if type(data_inp_1) is str:
        d1orig = files_rw.read_sp3(data_inp_1, epoch_as_pd_index=True)
    else:
        d1orig = data_inp_1.copy(True)
        try:
            d1orig.name = data_inp_1.name
        except:
            d1orig.name = "no_name"
        try:
            d1orig.path = data_inp_1.path
        except:
            d1orig.path = "no_path"
        try:
            d1orig.filename = data_inp_1.filename
        except:
            d1orig.filename = "no_filename"

    if type(data_inp_2) is str:
        d2orig = files_rw.read_sp3(data_inp_2, epoch_as_pd_index=True)
    else:
        d2orig = data_inp_2.copy(True)
        try:
            d2orig.name = data_inp_2.name
        except:
            d2orig.name = "no_name"
        try:
            d2orig.path = data_inp_2.path
        except:
            d2orig.path = "no_path"
        try:
            d2orig.filename = data_inp_2.filename
        except:
            d2orig.filename = "no_filename"

    #### NB : It has been decided with GM that the index of a SP3 dataframe
    ####      will be integers, not epoch datetime anymore
    ####      BUT here, for legacy reasons, the index has to be datetime

    if isinstance(d1orig.index[0], (int, np.integer)):
        d1orig.set_index("epoch", inplace=True)

    if isinstance(d2orig.index[0], (int, np.integer)):
        d2orig.set_index("epoch", inplace=True)

    diff_sat_stk = []

    # This block is for removing null values
    if clean_null_values:
        if clean_null_values == "all":
            all_or_any = np.all
        elif clean_null_values == "any":
            all_or_any = np.any
        else:
            all_or_any = np.all

        xyz_lst = ["x", "y", "z"]

        d1_null_bool = all_or_any(np.isclose(d1orig[xyz_lst], 0.0), axis=1)
        d2_null_bool = all_or_any(np.isclose(d2orig[xyz_lst], 0.0), axis=1)

        d1 = d1orig[np.logical_not(d1_null_bool)]
        d2 = d2orig[np.logical_not(d2_null_bool)]

        if np.any(d1_null_bool) or np.any(d2_null_bool):
            sat_nul = utils.join_improved(" ", *list(set(d1orig[d1_null_bool]["prn"])))
            log.warning("Null values contained in SP3 files : ")
            log.warning(
                "f1: %s %s",
                np.sum(d1_null_bool),
                utils.join_improved(" ", *list(set(d1orig[d1_null_bool]["prn"]))),
            )
            log.warning(
                "f2: %s %s",
                np.sum(d2_null_bool),
                utils.join_improved(" ", *list(set(d2orig[d2_null_bool]["prn"]))),
            )
        else:
            sat_nul = []

    else:
        d1 = d1orig.copy()
        d2 = d2orig.copy()

    d1sys_grp = d1.groupby("sys")
    d2sys_grp = d2.groupby("sys")

    for sysuse in sys_used_list:
        d1sys = d1sys_grp.get_group(sysuse)  # d1sys = d1[d1['sys'] == sysuse]
        d2sys = d2sys_grp.get_group(sysuse)  # d2sys = d2[d2['sys'] == sysuse]

        # checking if the data correspond to the step
        bool_step1 = np.mod((d1sys.index - np.min(d1.index)).seconds, step_data) == 0
        bool_step2 = np.mod((d2sys.index - np.min(d2.index)).seconds, step_data) == 0

        d1win = d1sys[bool_step1]
        d2win = d2sys[bool_step2]

        # find common sats and common epochs
        prni_set = sorted(list(set(d1win["prni"]).intersection(set(d2win["prni"]))))
        epoc_set = sorted(list(set(d1win.index).intersection(set(d2win.index))))

        # if special selection of sats, then apply it
        # (it is late and this selection is incredibely complicated ...)
        if np.any([True if sysuse in e else False for e in prn_used_list]):
            # first find the selected sats for the good constellation
            prni_used_select_list = [int(e[1:]) for e in prn_used_list if sysuse in e]
            # and apply it
            prni_set = sorted(
                list(set(prni_set).intersection(set(prni_used_select_list)))
            )

        d1win_prni_grp = d1win.groupby("prni")
        d2win_prni_grp = d2win.groupby("prni")

        for prni in prni_set:
            # First research : find corresponding epoch for the SV
            # this one is sufficent if there is no gaps (e.g. with 0.00000) i.e.
            # same nb of obs in the 2 files
            # NB : .reindex() is smart, it fills the DataFrame
            # with NaN
            try:
                d1prni_orig = d1win_prni_grp.get_group(prni).reindex(epoc_set)
                # d1win[d1win['prni'] == prni].reindex(epoc_set)
                d2prni_orig = d2win_prni_grp.get_group(prni).reindex(epoc_set)
                # d2win[d2win['prni'] == prni].reindex(epoc_set)
            except Exception as exce:
                log.info("ERR : Unable to re-index with an unique epoch")
                log.info(
                    "      are you sure there is no multiple-defined epochs for the same sat ?"
                )
                log.info(
                    "      it happens e.g. when multiple ACs are in the same DataFrame "
                )
                log.info(
                    "TIP : Filter the input Dataframe before calling this fct with"
                )
                log.info("      DF = DF[DF['AC'] == 'gbm']")

                dtmp1 = d1orig[d1orig["prni"] == prni]
                dtmp2 = d2orig[d2orig["prni"] == prni]

                dupli1 = np.sum(dtmp1.duplicated(["epoch", "prn"]))
                dupli2 = np.sum(dtmp2.duplicated(["epoch", "prn"]))

                log.info(
                    "FWIW: duplicated epoch/sat in DF1 & DF2: %s %s", dupli1, dupli2
                )

                raise exce

            # Second research, it is a security in case of gap
            # This step is useless, because .reindex() will fill the DataFrame
            # with NaN
            if len(d1prni_orig) != len(d2prni_orig):
                log.info(
                    "different epochs nbr for SV %s %s %s",
                    prni,
                    len(d1prni_orig),
                    len(d2prni_orig),
                )
                epoc_prni_set = sorted(
                    list(set(d1prni_orig.index).intersection(set(d2prni_orig.index)))
                )
                d1prni = d1prni_orig.loc[epoc_prni_set]
                d2prni = d2prni_orig.loc[epoc_prni_set]
            else:
                d1prni = d1prni_orig
                d2prni = d2prni_orig

            p1 = d1prni[["x", "y", "z"]]
            p2 = d2prni[["x", "y", "z"]]

            # Start ECEF => ECI
            if convert_ecef_eci:
                # Backup because the columns xyz will be reaffected
                # D1sv_bkp = d1prni.copy()
                # D2sv_bkp = d2prni.copy()

                p1b = conv.ecef2eci(
                    np.array(p1),
                    conv.dt_gpstime2dt_utc(p1.index.to_pydatetime(), out_array=True),
                )
                p2b = conv.ecef2eci(
                    np.array(p2),
                    conv.dt_gpstime2dt_utc(p2.index.to_pydatetime(), out_array=True),
                )

                d1prni[["x", "y", "z"]] = p1b
                d2prni[["x", "y", "z"]] = p2b

                p1 = d1prni[["x", "y", "z"]]
                p2 = d2prni[["x", "y", "z"]]
            # End ECEF => ECI

            if not rtn_output:
                # Compatible with the documentation +
                # empirically tested with OV software
                # it is  p1 - p2 (and not p2 - p1)
                delta_p = p1 - p2

                diff_sat = delta_p.copy()
                diff_sat.columns = ["dx", "dy", "dz"]

            else:
                rnorm = np.linalg.norm(p1, axis=1)

                from geodezyx.utils_xtra import pandas_utils

                vx = pandas_utils.diff_pandas(d1prni, "x", use_np_diff=True)
                vy = pandas_utils.diff_pandas(d1prni, "y", use_np_diff=True)
                vz = pandas_utils.diff_pandas(d1prni, "z", use_np_diff=True)

                v = pd.concat((vx, vy, vz), axis=1)
                v.columns = ["vx", "vy", "vz"]

                r = p1.divide(rnorm, axis=0)
                r.columns = ["xnorm", "ynorm", "znorm"]

                h = pd.DataFrame(np.cross(r, v), columns=["hx", "hy", "hz"])
                hnorm = np.linalg.norm(h, axis=1)

                c = h.divide(hnorm, axis=0)
                c.columns = ["hxnorm", "hynorm", "hznorm"]

                i = pd.DataFrame(np.cross(c, r), columns=["ix", "iy", "iz"])

                r_ar = np.array(r)
                i_ar = np.array(i)
                c_ar = np.array(c)

                # r_ar[1]
                beta = np.stack((r_ar, i_ar, c_ar), axis=1)

                # Compatible with the documentation +
                # empirically tested with OV software
                # it is  p1 - p2 (and not p2 - p1)
                delta_p = p1 - p2

                # Final determination
                astk = []

                for i in range(len(delta_p)):
                    a = np.dot(beta[i, :, :], np.array(delta_p)[i])
                    astk.append(a)

                diff_sat = pd.DataFrame(
                    np.vstack(astk), index=p1.index, columns=["dr", "dt", "dn"]
                )

            diff_sat = diff_sat * conv_coef  # metrer conversion

            diff_sat["sys"] = [sysuse] * len(diff_sat.index)
            diff_sat["prni"] = [prni] * len(diff_sat.index)
            diff_sat["prn"] = [sysuse + str(prni).zfill(2)] * len(diff_sat.index)

            diff_sat_stk.append(diff_sat)

    diff_sat_all = pd.concat(diff_sat_stk)
    date = diff_sat.index[0]

    # Attribute definition
    if rtn_output:
        diff_sat_all.frame_type = "RTN"

        # Pandas donesn't manage well iterable as attribute
        # So, it is separated
        diff_sat_all.frame_col_name1 = "dr"
        diff_sat_all.frame_col_name2 = "dt"
        diff_sat_all.frame_col_name3 = "dn"

    else:
        # Pandas donesn't manage well iterable as attribute
        # So, it is separated
        diff_sat_all.frame_col_name1 = "dx"
        diff_sat_all.frame_col_name2 = "dy"
        diff_sat_all.frame_col_name3 = "dz"

        if convert_ecef_eci:
            diff_sat_all.frame_type = "ECI"
        else:
            diff_sat_all.frame_type = "ECEF"

    # Name definitions
    if name1:
        diff_sat_all.name1 = name1
    else:
        diff_sat_all.name1 = d1orig.name

    if name2:
        diff_sat_all.name2 = name2
    else:
        diff_sat_all.name2 = d2orig.name

    diff_sat_all.filename1 = d1orig.filename
    diff_sat_all.filename2 = d2orig.filename

    diff_sat_all.path1 = d1orig.path
    diff_sat_all.path2 = d2orig.path

    diff_sat_all.name = " ".join(
        (
            "Orbits comparison (" + diff_sat_all.frame_type + ") b/w",
            diff_sat_all.name1,
            "(ref.) and",
            diff_sat_all.name2,
            ",",
            date.strftime("%Y-%m-%d"),
            ", doy",
            str(conv.dt2doy(date)),
        )
    )

    if return_sat_null:
        return diff_sat_all, sat_nul
    else:
        return diff_sat_all


def compar_orbit_plot(
    diff_sat_all_df_in,
    save_plot=False,
    save_plot_dir="",
    save_plot_name="auto",
    save_plot_name_suffix=None,
    save_plot_ext=(".pdf", ".png", ".svg"),
    yaxis_limit=None,
    yaxis_label_unit="m",
):
    """
    General description

    Parameters
    ----------
    diff_sat_all_df_in : DataFrame
        a DataFrame produced by compar_orbit

    yaxis_limit : 3-tuple iterable or 2-element tuple
        force the y axis limits. must look like
        [(ymin_r,ymax_r),(ymin_t,ymax_t),(ymin_n,ymax_n)]
        to control all the axis independely
        OR
        (ymin,ymax)
        to set all th axis at the same limits

    Returns
    -------
    the Figure and the 3 Axes if no save is asked
    export path (str) if save is asked
    but plot a plot anyway
    """

    fig, [axr, axt, axn] = plt.subplots(3, 1, sharex="all")

    satdispo = natsort.natsorted(list(set(diff_sat_all_df_in["prn"])))

    symb_stk = []

    cm = plt.get_cmap("viridis")
    num_colors = len(satdispo)
    colors = [cm(1.0 * i / num_colors) for i in range(num_colors)]

    # Pandas donesn't manage well iterable as attribute
    # So, it is separated
    try:
        col_name0 = diff_sat_all_df_in.frame_col_name1
        col_name1 = diff_sat_all_df_in.frame_col_name2
        col_name2 = diff_sat_all_df_in.frame_col_name3
    except:
        col_name0 = diff_sat_all_df_in.columns[0]
        col_name1 = diff_sat_all_df_in.columns[1]
        col_name2 = diff_sat_all_df_in.columns[2]

    for satuse, color in zip(satdispo, colors):
        diffuse = diff_sat_all_df_in[diff_sat_all_df_in["prn"] == satuse]

        time = diffuse.index
        r = diffuse[col_name0]
        t = diffuse[col_name1]
        n = diffuse[col_name2]

        # fig.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')

        symb = axr.plot(time, r, label=satuse, c=color)
        axt.plot(time, t, label=satuse, c=color)
        axn.plot(time, n, label=satuse, c=color)

        symb_stk.append(symb[0])

        fig.autofmt_xdate()

    ylabuni = " (" + yaxis_label_unit + ")"

    if diff_sat_all_df_in.frame_type == "RTN":
        axr.set_ylabel("Radial diff." + ylabuni)
        axt.set_ylabel("Transverse diff." + ylabuni)
        axn.set_ylabel("Normal diff." + ylabuni)

    else:
        axr.set_ylabel(diff_sat_all_df_in.frame_type + " X diff." + ylabuni)
        axt.set_ylabel(diff_sat_all_df_in.frame_type + " Y diff." + ylabuni)
        axn.set_ylabel(diff_sat_all_df_in.frame_type + " Z diff." + ylabuni)

    y_formatter = matplotlib.ticker.ScalarFormatter(useOffset=False)
    axr.yaxis.set_major_formatter(y_formatter)
    axt.yaxis.set_major_formatter(y_formatter)
    axn.yaxis.set_major_formatter(y_formatter)

    if yaxis_limit and len(yaxis_limit) == 3:  ### indep. axis limit
        axr.set_ylim(yaxis_limit[0])
        axt.set_ylim(yaxis_limit[1])
        axn.set_ylim(yaxis_limit[2])
    elif yaxis_limit and len(yaxis_limit) == 2:
        axr.set_ylim(yaxis_limit)
        axt.set_ylim(yaxis_limit)
        axn.set_ylim(yaxis_limit)
    else:
        pass

    import matplotlib.dates as mdates

    fig.fmt_xdata = mdates.DateFormatter("%Y-%m-%d")

    lgd = fig.legend(
        tuple(symb_stk), satdispo, loc="lower center", ncol=8, columnspacing=1
    )

    fig.set_size_inches(8.27, 11.69)
    plt.suptitle(diff_sat_all_df_in.name)
    plt.tight_layout()
    plt.subplots_adjust(top=0.95)
    plt.subplots_adjust(bottom=0.15)

    if save_plot:
        if save_plot_name == "auto":
            save_plot_name = "_".join(
                (
                    diff_sat_all_df_in.name1,
                    diff_sat_all_df_in.name2,
                    diff_sat_all_df_in.index.min().strftime("%Y-%m-%d"),
                )
            )

        if save_plot_name_suffix:
            save_plot_name = save_plot_name + "_" + save_plot_name_suffix

        for ext in save_plot_ext:
            save_plot_path = os.path.join(save_plot_dir, save_plot_name)
            plt.savefig(save_plot_path + ext)
            return_val = save_plot_path

    else:
        return_val = fig, (axr, axt, axn)

    return return_val


def compar_orbit_table(diff_sat_all_df_in, rms_style="natural", light_tab=False):
    """
    Generate a table with statistical indicators for an orbit comparison
    (RMS mean, standard dev, ...)

    Parameters
    ----------
    diff_sat_all_df_in : Pandas DataFrame
        a DataFrame produced by compar_orbit

    rms_style : str
        'natural': use the natural definition of the RMS
        'GRGS': RMS calc based on the GRGS definition of the RMS (OV help)
        is actually the standard deviation
        'kouba': RMS as defined in Kouba et al. 1994, p75
        using the degree of freedom (3*Nobs - 7)

    light_tab : bool
        produce a table with only RMS, with min/max/arithmetic instead

    Returns
    -------
    Compar_tab_out : DataFrame
        Statistical results of the comparison

    Note
    ----
    you can pretty print the output DataFrame using tabular module
    here is a template:

    >>> from tabulate import tabulate
    >>> print(tabulate(compar_table,headers="keys",floatfmt=".4f"))
    """

    sat_list = utils.uniq_and_sort(diff_sat_all_df_in["prn"])

    # Pandas donesn't manage well iterable as attribute
    # So, it is separated
    try:
        col_name0 = diff_sat_all_df_in.frame_col_name1
        col_name1 = diff_sat_all_df_in.frame_col_name2
        col_name2 = diff_sat_all_df_in.frame_col_name3
    except:
        col_name0 = diff_sat_all_df_in.columns[0]
        col_name1 = diff_sat_all_df_in.columns[1]
        col_name2 = diff_sat_all_df_in.columns[2]

    rms_stk = []

    for sat in sat_list:
        diffwork = utils.df_sel_val_in_col(diff_sat_all_df_in, "prn", sat)

        if rms_style == "natural":
            rms_a = stats.rms_mean(diffwork[col_name0])
            rms_b = stats.rms_mean(diffwork[col_name1])
            rms_c = stats.rms_mean(diffwork[col_name2])
        elif rms_style == "GRGS":
            rms_a = stats.rms_mean(diffwork[col_name0] - diffwork[col_name0].mean())
            rms_b = stats.rms_mean(diffwork[col_name1] - diffwork[col_name1].mean())
            rms_c = stats.rms_mean(diffwork[col_name2] - diffwork[col_name2].mean())
        elif rms_style == "kouba":
            rms_a = stats.rms_mean_kouba(diffwork[col_name0])
            rms_b = stats.rms_mean_kouba(diffwork[col_name1])
            rms_c = stats.rms_mean_kouba(diffwork[col_name2])

        rms_3d = np.sqrt(rms_a**2 + rms_b**2 + rms_c**2)

        min_a = diffwork[col_name0].min()
        min_b = diffwork[col_name1].min()
        min_c = diffwork[col_name2].min()

        max_a = diffwork[col_name0].max()
        max_b = diffwork[col_name1].max()
        max_c = diffwork[col_name2].max()

        mean_a = diffwork[col_name0].mean()
        mean_b = diffwork[col_name1].mean()
        mean_c = diffwork[col_name2].mean()

        if light_tab:
            rms_stk.append([rms_a, rms_b, rms_c, rms_3d])
        else:
            rms_stk.append(
                [
                    rms_a,
                    rms_b,
                    rms_c,
                    rms_3d,
                    min_a,
                    max_a,
                    mean_a,
                    min_b,
                    max_b,
                    mean_b,
                    min_c,
                    max_c,
                    mean_c,
                ]
            )

    #################################
    # ALL SATS
    if rms_style == "natural":
        rms_a = stats.rms_mean(diff_sat_all_df_in[col_name0])
        rms_b = stats.rms_mean(diff_sat_all_df_in[col_name1])
        rms_c = stats.rms_mean(diff_sat_all_df_in[col_name2])
        rms_3d = np.sqrt(rms_a**2 + rms_b**2 + rms_c**2)
    elif rms_style == "GRGS":
        rms_a = stats.rms_mean(
            diff_sat_all_df_in[col_name0] - diff_sat_all_df_in[col_name0].mean()
        )
        rms_b = stats.rms_mean(
            diff_sat_all_df_in[col_name1] - diff_sat_all_df_in[col_name1].mean()
        )
        rms_c = stats.rms_mean(
            diff_sat_all_df_in[col_name2] - diff_sat_all_df_in[col_name2].mean()
        )
        rms_3d = np.sqrt(rms_a**2 + rms_b**2 + rms_c**2)
    elif rms_style == "kouba":
        rms_a = stats.rms_mean_kouba(diff_sat_all_df_in[col_name0])
        rms_b = stats.rms_mean_kouba(diff_sat_all_df_in[col_name1])
        rms_c = stats.rms_mean_kouba(diff_sat_all_df_in[col_name2])
        rms_3d = np.sqrt(rms_a**2 + rms_b**2 + rms_c**2)

    min_a = diff_sat_all_df_in[col_name0].min()
    min_b = diff_sat_all_df_in[col_name1].min()
    min_c = diff_sat_all_df_in[col_name2].min()

    max_a = diff_sat_all_df_in[col_name0].max()
    max_b = diff_sat_all_df_in[col_name1].max()
    max_c = diff_sat_all_df_in[col_name2].max()

    mean_a = diff_sat_all_df_in[col_name0].mean()
    mean_b = diff_sat_all_df_in[col_name1].mean()
    mean_c = diff_sat_all_df_in[col_name2].mean()

    if light_tab:
        rms_stk.append([rms_a, rms_b, rms_c, rms_3d])
    else:
        rms_stk.append(
            [
                rms_a,
                rms_b,
                rms_c,
                rms_3d,
                min_a,
                max_a,
                mean_a,
                min_b,
                max_b,
                mean_b,
                min_c,
                max_c,
                mean_c,
            ]
        )

        # ALL SATS
    #################################

    if diff_sat_all_df_in.frame_type == "RTN":
        if light_tab:
            cols_nam = ["rmsR", "rmsT", "rmsN", "rms3D"]
        else:
            cols_nam = [
                "rmsR",
                "rmsT",
                "rmsN",
                "rms3D",
                "minR",
                "maxR",
                "meanR",
                "minT",
                "maxT",
                "meanT",
                "minN",
                "maxN",
                "meanN",
            ]

    else:
        if light_tab:
            cols_nam = ["rmsX", "rmsY", "rmsZ", "rms3D"]
        else:
            cols_nam = [
                "rmsX",
                "rmsY",
                "rmsZ",
                "rms3D",
                "minX",
                "maxX",
                "meanX",
                "minY",
                "maxY",
                "meanY",
                "minZ",
                "maxZ",
                "meanZ",
            ]

    compar_tab_out = pd.DataFrame(rms_stk, index=sat_list + ["ALL"], columns=cols_nam)

    return compar_tab_out


def compar_orbit_frontend(data_df1, data_df2, ac1, ac2, sats_used_list=["G"]):
    K = compar_orbit(
        data_df1[data_df1["ac"] == ac1],
        data_df2[data_df2["ac"] == ac2],
        sats_used_list=sats_used_list,
    )
    compar_orbit_plot(K)
    return K


def compar_clock(df_clk_inp_1, df_clk_inp_2, col_name="name", bias_col_name="bias"):
    """
    Compares 2 GNSS clock bias DataFrames (from .clk), to a
    statistics table (with compar_clock_table)


    Parameters
    ----------
    df_clk_inp_1 & df_clk_inp_2 : DataFrame
        Clock DataFrame provided by files_rw.read_clk()

    Returns
    -------
    df_clk_diff : DataFrame
        Clock bias difference DataFrame
    """
    df1idx = df_clk_inp_1.set_index([col_name, "epoch"])
    df1idx.sort_index(inplace=True)

    df2idx = df_clk_inp_2.set_index([col_name, "epoch"])
    df2idx.sort_index(inplace=True)

    i1 = df1idx.index
    i2 = df2idx.index

    iinter = i1.intersection(i2)
    iinter = iinter.sort_values()

    df_diff_bias = df1idx.loc[iinter][bias_col_name] - df2idx.loc[iinter][bias_col_name]

    df_clk_diff = df1idx.loc[iinter].copy()
    df_clk_diff[bias_col_name] = df_diff_bias
    if "ac" in df_clk_diff.columns:
        df_clk_diff.drop("ac", axis=1, inplace=True)
    else:
        df_clk_diff.drop("AC", axis=1, inplace=True)
    df_clk_diff.rename({bias_col_name: bias_col_name + "_diff"}, inplace=True, axis=1)

    # Name definitions
    if "ac" in df_clk_inp_1.columns:
        df_clk_diff["name1"] = df_clk_inp_1.ac.values[0]
    if "AC" in df_clk_inp_1.columns:
        df_clk_diff["name1"] = df_clk_inp_1.AC.values[0]

    if "ac" in df_clk_inp_1.columns:
        df_clk_diff["name2"] = df_clk_inp_2.ac.values[0]
    if "AC" in df_clk_inp_1.columns:
        df_clk_diff["name2"] = df_clk_inp_2.AC.values[0]

    return df_clk_diff


def compar_clock_table(df_clk_diff_in, col_name="name", bias_Col_name="bias_diff"):
    """
    Generate a table with statistical indicators for a clock comparison
    (RMS mean, standard dev, ...)

    Parameters
    ----------
    df_clk_diff_in : DataFrame
        Clock bias difference DataFrame (from compar_clock)

    Returns
    -------
    DFcompar_out : DataFrame
        Statistical results of the comparison.

    """

    df_diff_grp = df_clk_diff_in.groupby(col_name)[bias_Col_name]

    smin = df_diff_grp.min().rename("min", inplace=True)
    smax = df_diff_grp.max().rename("max", inplace=True)
    smean = df_diff_grp.mean().rename("mean", inplace=True)
    sstd = df_diff_grp.std().rename("std", inplace=True)
    srms = df_diff_grp.apply(stats.rms_mean).rename("rms", inplace=True)

    df_compar_out = pd.concat([smin, smax, smean, sstd, srms], axis=1)
    df_compar_out.reset_index()

    return df_compar_out


def compar_clk_plot(
    diff_sat_all_df_in,
    save_plot=False,
    save_plot_dir="",
    save_plot_name="auto",
    save_plot_name_suffix=None,
    save_plot_ext=(".pdf", ".png", ".svg"),
    yaxis_limit=None,
    yaxis_label_unit="psec",
    col_name="name",
    bias_Col_name="bias",
):
    """
    General description

    Parameters
    ----------
    diff_sat_all_df_in: DataFrame
        a DataFrame produced by compar_clk

    yaxis_limit: 3-tuple iterable or 2-element tuple
        force the y axis limits. must look like
        (ymin,ymax) to set all th axis at the same limits

    col_name: Normally the name of the column with the sat names
    bias_Col_name: The column with the clk values
    Default: 'bias'

    Returns
    -------
    the Figure and the 3 Axes if no save is asked
    export path (str) if save is asked
    but plot a plot anyway
    """

    fig, axr = plt.subplots(1, 1, sharex="all")
    diff_sat_all_df_in = diff_sat_all_df_in.reset_index()
    satdispo = natsort.natsorted(list(set(diff_sat_all_df_in[col_name])))
    # satdispo = natsort.natsorted(list(set(diff_sat_all_df_in['sat'])))

    symb_stk = []

    cm = plt.get_cmap("viridis")
    num_colors = len(satdispo)
    colors = [cm(1.0 * i / num_colors) for i in range(num_colors)]

    date = conv.numpy_dt2dt(diff_sat_all_df_in.epoch.values[0])
    diff_sat_all_df_in.name = " ".join(
        (
            "Clock comparison  b/w",
            diff_sat_all_df_in.name1.values[0],
            "(ref.) and",
            diff_sat_all_df_in.name2.values[0],
            ",",
            date.strftime("%Y-%m-%d"),
        )
    )
    # Pandas donesn't manage well iterable as attribute
    # So, it is separated
    try:
        col_name0 = diff_sat_all_df_in.frame_col_name1
        col_name1 = diff_sat_all_df_in.frame_col_name2
        col_name2 = diff_sat_all_df_in.frame_col_name3
    except:
        col_name0 = diff_sat_all_df_in.columns[0]
        col_name1 = diff_sat_all_df_in.columns[1]
        col_name2 = diff_sat_all_df_in.columns[2]

    for satuse, color in zip(satdispo, colors):
        diffuse = diff_sat_all_df_in[diff_sat_all_df_in[col_name] == satuse]

        time = diffuse.epoch
        r = diffuse[bias_Col_name + "_diff"] * 10**12

        # fig.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')

        Symb = axr.plot(time, r, label=satuse, c=color)

        symb_stk.append(Symb[0])

        fig.autofmt_xdate()

    ylabuni = " (" + yaxis_label_unit + ")"

    axr.set_ylabel("Bias Diff." + ylabuni)

    y_formatter = matplotlib.ticker.ScalarFormatter(useOffset=False)
    axr.yaxis.set_major_formatter(y_formatter)

    if yaxis_limit and len(yaxis_limit) == 3:  ### indep. axis limit
        axr.set_ylim(yaxis_limit[0])

    elif yaxis_limit and len(yaxis_limit) == 2:
        axr.set_ylim(yaxis_limit)

    else:
        pass

    import matplotlib.dates as mdates

    fig.fmt_xdata = mdates.DateFormatter("%Y-%m-%d")

    lgd = fig.legend(
        tuple(symb_stk), satdispo, loc="lower center", ncol=8, columnspacing=1
    )

    fig.set_size_inches(8.27, 11.69)
    plt.suptitle(diff_sat_all_df_in.name)
    plt.tight_layout()
    plt.subplots_adjust(top=0.95)
    plt.subplots_adjust(bottom=0.15)

    if save_plot:
        if save_plot_name == "auto":
            save_plot_name = (
                diff_sat_all_df_in.name1.values[0]
                + "_"
                + diff_sat_all_df_in.name2.values[0]
                + "_"
                + date.strftime("%Y-%m-%d")
            )

        if save_plot_name_suffix:
            save_plot_name = save_plot_name + "_" + save_plot_name_suffix

        for ext in save_plot_ext:
            save_plot_path = os.path.join(save_plot_dir, save_plot_name)
            plt.savefig(save_plot_path + ext)
            return_val = save_plot_path

    else:
        return_val = fig, (axr)

    return return_val


def compar_sinex(
    snx1,
    snx2,
    stat_select=None,
    invert_select=False,
    out_means_summary=True,
    out_meta=True,
    out_dataframe=True,
    manu_wwwwd=None,
):

    if type(snx1) is str:
        week1 = utils.split_improved(os.path.basename(snx1), "_", ".")[:]
        week2 = utils.split_improved(os.path.basename(snx2), "_", ".")[:]
        if week1 != week2:
            log.warning(
                "Dates of 2 input files are differents !!! It might be very bad !!! %s %s",
                week1,
                week2,
            )
        else:
            wwwwd = week1
        d1 = files_rw.read_sinex(snx1, True)
        d2 = files_rw.read_sinex(snx2, True)
    else:
        log.warning(
            "WARN : you are giving the SINEX input as a DataFrame, wwwwd has to be given manually using manu_wwwwd"
        )
        d1 = snx1
        d2 = snx2

    if manu_wwwwd:
        wwwwd = manu_wwwwd

    stat_common = set(d1["STAT"]).intersection(set(d2["STAT"]))

    if stat_select:

        stat_common_init = list(stat_common)

        if invert_select:
            select_fct = lambda x: not x
        else:
            select_fct = lambda x: x

        if type(stat_select) is str:
            stat_common = [
                sta
                for sta in stat_common_init
                if select_fct(re.search(stat_select, sta))
            ]
        elif utils.is_iterable(stat_select):
            stat_common = [
                sta for sta in stat_common_init if select_fct(sta in stat_select)
            ]
        else:
            log.warning("WARN : check type of stat_select")

    d1_common = (
        d1[d1["STAT"].isin(stat_common)].sort_values("STAT").reset_index(drop=True)
    )
    d2_common = (
        d2[d2["STAT"].isin(stat_common)].sort_values("STAT").reset_index(drop=True)
    )

    ddiff = pd.DataFrame()
    ddiff = ddiff.assign(STAT=d1_common["STAT"])

    #### XYZ Part
    for xyz in ("x", "y", "z"):

        dif = pd.to_numeric((d2_common[xyz] - d1_common[xyz]))

        ddiff = ddiff.assign(xyz=dif)
        ddiff = ddiff.rename(columns={"xyz": xyz})

    D3D = np.sqrt(
        (ddiff["x"] ** 2 + ddiff["y"] ** 2 + ddiff["z"] ** 2).astype("float64")
    )

    ddiff = ddiff.assign(d3D_xyz=D3D)

    ### ENU Part
    E, N, U = [], [], []
    enu_stk = []

    for (_, l1), (_, l2) in zip(d1_common.iterrows(), d2_common.iterrows()):
        enu = conv.xyz2enu(l1["x"], l1["y"], l1["z"], l2["x"], l2["y"], l2["z"])
        enu_stk.append(np.array(enu))

    if len(enu_stk) == 0:
        E, N, U = np.array([]), np.array([]), np.array([])
    else:
        ENU = np.hstack(enu_stk)
        E, N, U = ENU[0, :], ENU[1, :], ENU[2, :]

    D2D = np.sqrt((E**2 + N**2).astype("float64"))
    D3D = np.sqrt((E**2 + N**2 + U**2).astype("float64"))

    ddiff = ddiff.assign(e=E)
    ddiff = ddiff.assign(n=N)
    ddiff = ddiff.assign(u=U)
    ddiff = ddiff.assign(d2D_enu=D2D)
    ddiff = ddiff.assign(d3D_enu=D3D)

    #    E,N,U    = conv.xyz2enu((X,Y,Z,x0,y0,z0))
    #    E,N,U    = conv.xyz2enu((X,Y,Z,x0,y0,z0))

    if out_dataframe:
        out_meta = True

    if not out_means_summary:
        log.info("INFO : this is not used operationally and it can be improved")
        return ddiff
    else:
        output = []

        col_names = ("x", "y", "z", "d3D_xyz", "e", "n", "u", "d2D_enu", "d3D_enu")

        for xyz in col_names:
            output.append(stats.rms_mean(ddiff[xyz]))
        for xyz in col_names:
            output.append(np.nanmean(ddiff[xyz]))
        for xyz in col_names:
            output.append(np.nanstd(ddiff[xyz]))

        if out_meta:
            nstat = len(stat_common)
            week = int(wwwwd[:4])
            day = int(wwwwd[4:])
            output = [week, day, nstat] + output

        if not out_dataframe:
            return tuple(output)
        else:

            output_DF = pd.DataFrame(output).transpose()

            output_DF.columns = [
                "week",
                "dow",
                "nbstat",
                "x_rms",
                "y_rms",
                "z_rms",
                "d3D_xyz_rms",
                "e_rms",
                "n_rms",
                "u_rms",
                "d2D_enu_rms",
                "d3D_enu_rms",
                "x_ari",
                "y_ari",
                "z_ari",
                "d3D_xyz_ari",
                "e_ari",
                "n_ari",
                "u_ari",
                "d2D_enu_ari",
                "d3D_enu_ari",
                "x_ari",
                "y_std",
                "z_std",
                "d3D_xyz_std",
                "e_ari",
                "n_std",
                "u_std",
                "d2D_enu_std",
                "d3D_enu_std",
            ]

            return output_DF


#   ____       _     _ _     _____        _        ______
#  / __ \     | |   (_) |   |  __ \      | |      |  ____|
# | |  | |_ __| |__  _| |_  | |  | | __ _| |_ __ _| |__ _ __ __ _ _ __ ___   ___  ___
# | |  | | '__| '_ \| | __| | |  | |/ _` | __/ _` |  __| '__/ _` | '_ ` _ \ / _ \/ __|
# | |__| | |  | |_) | | |_  | |__| | (_| | || (_| | |  | | | (_| | | | | | |  __/\__ \
#  \____/|_|  |_.__/|_|\__| |_____/ \__,_|\__\__,_|_|  |_|  \__,_|_| |_| |_|\___||___/
#

### Orbit DataFrames


def orb_df_velocity_calc(df_orb_in, drop_nan=False):
    """
    Compute the velocity of satellites from a DFOrb dataframe
    (differentiate the position)

    Parameters
    ----------
    df_orb_in : Pandas DataFrame
        an Orbit DataFrame.
    drop_nan : bool, optional
        Remove the nan values
        (the first ones, since it is a numerical differentiation).
        The default is False.
        It is recommended to keep the NaN values, thus the output will keep
        same size and index as input

    Returns
    -------
    df_vel : Pandas DataFrame
        an Orbit DataFrame with velocities (vx, vy ,vz columns).

    """

    dfgrp = df_orb_in.groupby("prn")

    dfprn_stk = []
    for prn, dfprn in dfgrp:
        for coord in ["x", "y", "z"]:
            dcoord = dfprn[coord].diff()
            dtime = np.float64(dfprn["epoch"].diff()) * 10**-9
            # timedelta in ns per defalut
            dfprn["v" + coord] = dcoord / dtime
        dfprn_stk.append(dfprn)

    df_vel = pd.concat(dfprn_stk)

    df_vel.sort_index(inplace=True)

    if drop_nan:
        df_vel.dropna(inplace=True)

    return df_vel


def beta_sun_ra_dec(sun_dec, sun_ra, sat_i, sat_o_lan):
    """
    Compute beta angle based on Sun's right ascension and declination
    Angles are in radians

    Parameters
    ----------
    sun_dec : float
        Sun's declination.
    sun_ra : float
        Sun's right ascension.
    sat_i : float
        Satellite's inclination.
    sat_o_lan : float
        Satellite's Longitude of the ascending node.

    Returns
    -------
    beta : float
        beta angle (in radians).

    Source
    ------
    `Satellites: de Kepler au GPS`, Michel Capderou (2012)

    simpler formula here:
        https://www.fxsolver.com/browse/formulas/Beta+Angle


    Note
    ----
    you can use ``pyorbital.astronomy.sun_ra_dec()`` to compute ``sun_dec``
    and ``sun_ra``

    If so, Sun's right ascension and declination computation polynoms are
    based on: `Astronomical algorithms`, Jean Meeus (1st edition, 1991)
    """
    beta = np.arcsin(
        np.cos(sun_dec) * np.sin(sat_i) * np.sin(sat_o_lan - sun_ra)
        + np.sin(sun_dec) * np.cos(sat_i)
    )
    return beta


def beta_sun_eclip_long(sun_ecl_long, sat_o_lan, sat_i, earth_i):
    """
    Compute beta angle based on Sun's Ecliptic longitude
    Angles are in radians

    Parameters
    ----------
    sun_ecl_long : float
        Sun's Ecliptic longitude.
    sat_i : float
        Satellite's inclination.
    sat_o_lan : float
        Satellite's Longitude of the ascending node.
    earth_i : float
        Earth's inclination.

    Returns
    -------
    beta : float
        beta angle (in radians).

    Source
    ------
    `Calculation of the Eclipse Factor for Elliptical Satellite Orbits` NASA (1962)
    https://ntrs.nasa.gov/citations/19630000622
    `Computation of Eclipse Time for Low-Earth Orbiting Small Satellites` Sumanth R M (2019)
    https://commons.erau.edu/ijaaa/vol6/iss5/15/

    Simpler formula:
        https://en.wikipedia.org/wiki/Beta_angle

    Note
    ----
    you can use ``pyorbital.astronomy.sun_ecliptic_longitude()``
    to compute ``sun_ecl_long``

    If so, Sun's right ascension and declination computation polynoms are
    based on: `Astronomical algorithms`, Jean Meeus (1st edition, 1991)

    """
    p1 = np.cos(sun_ecl_long) * np.sin(sat_o_lan) * np.sin(sat_i)
    p2 = np.sin(sun_ecl_long) * np.cos(earth_i) * np.cos(sat_o_lan) * np.sin(sat_i)
    p3 = np.sin(sun_ecl_long) * np.sin(earth_i) * np.cos(sat_i)
    beta = np.arcsin(p1 - p2 + p3)
    return beta


def beta_angle_calc(
    DFOrb_in,
    calc_beta_sun_ra_dec=True,
    calc_beta_sun_eclip_long=True,
    beta_rad2deg=True,
):
    """
    Compute beta angle for GNSS satellite's orbits stored in an orbit
    DataFrame


    Parameters
    ----------
    DFOrb_in : Pandas DataFrame
        an Orbit DataFrame (ECEF frame).
    calc_beta_sun_ra_dec : bool, optional
        compute beta angle with Sun's right ascension and declination.
        The default is True.
    calc_beta_sun_eclip_long : bool, optional
        compute beta angle with Sun's Ecliptic longitude.
        The default is True.
    beta_rad2deg : bool, optional
        convert beta angle to degrees. The default is True.

    Returns
    -------
    df_out : Pandas DataFrame
        df_orb_in with a new 'beta' column.
    df_wrk : Pandas DataFrame
        intermediate values dataframe for debug.
        here, coordinates are in ECI frame
    """
    import pyorbital.astronomy

    #### convert in ECI
    df_eci = DFOrb_in.copy()
    df_eci[["x", "y", "z"]] = conv.ecef2eci(
        DFOrb_in[["x", "y", "z"]].values, DFOrb_in["epoch"].values
    )

    #### compute velocity
    df_wrk = orb_df_velocity_calc(df_eci, drop_nan=False)
    ##keep drop nan False, thus the DF will keep same size and index as input

    #### compute Kepler's parameters
    p = df_wrk[["x", "y", "z"]].values * 1000
    v = df_wrk[["vx", "vy", "vz"]].values * 1000

    kep_col = ["a", "ecc", "i", "o_peri", "o_lan", "m"]
    kep_params = kepler_gzyx.eci_2_kepler_elts(p, v, rad2deg=False)
    df_wrk[kep_col] = np.column_stack(kep_params)
    ######### COMPUTE BETA

    #### cosmetic changes
    if calc_beta_sun_ra_dec and calc_beta_sun_eclip_long:
        b1 = "1"
        b2 = "2"
    else:
        b1 = ""
        b2 = ""

    if beta_rad2deg:
        r2dfct = np.rad2deg
    else:
        r2dfct = lambda x: x

    ##### Beta computed based on sun declination / right ascension
    if calc_beta_sun_ra_dec:
        ##### sun_ra_dec output in RADIANS
        df_wrk[["sun_ra", "sun_dec"]] = np.column_stack(
            pyorbital.astronomy.sun_ra_dec(df_wrk["epoch"])
        )
        df_wrk["beta" + b1] = beta_sun_ra_dec(
            df_wrk["sun_dec"], df_wrk["sun_ra"], df_wrk["i"], df_wrk["o_lan"]
        ).apply(r2dfct)

    ##### Beta computed based on Ecliptic longitude of the sun
    if calc_beta_sun_eclip_long:
        ### sun_ecliptic_longitude output in RADIANS but NO MODULO !!!
        df_wrk["sun_ecl_long"] = np.mod(
            pyorbital.astronomy.sun_ecliptic_longitude(df_wrk["epoch"]), np.pi * 2
        )
        df_wrk["beta" + b2] = beta_sun_eclip_long(
            df_wrk["sun_ecl_long"], df_wrk["o_lan"], df_wrk["i"], np.deg2rad(23.45)
        ).apply(r2dfct)
    df_out = DFOrb_in.copy()
    df_out["beta"] = df_wrk["beta" + b1]

    return df_out, df_wrk


def orb_df_lagrange_interpolate(
    orb_df_inp, titrp, n=10, append_to_input_df=False, plot=False
):
    """
    High level function to interpolate an orbit DataFrame

    Parameters
    ----------
    orb_df_inp : DataFrame
        an Orbit DataFrame.
    titrp : iterable of datetime
        Epochs of the wished points.
    n : int, optional
        degree of the polynom. Better if even. The default is 10.
    append_to_input_df : bool, optional
        append the interpolated DF to the input DF. The default is False.
    plot : bool, optional
        Plot the values. For debug only. The default is False.

    Returns
    -------
    DForb_out : DataFrame
        Interpolated orbits.

    Tips
    ----
    Use conv.dt_range to generate the wished epochs range

    """
    df_orb_stk = []

    for sat, ac in itertools.product(
        orb_df_inp["prn"].unique(), orb_df_inp["ac"].unique()
    ):

        log.info("process %s %s", ac, sat)

        DForb_use = orb_df_inp[
            (orb_df_inp["prn"] == sat) & (orb_df_inp["ac"] == ac)
        ].copy()

        ### faster but anoying Future Waring
        # Tdata = np.array(DForb_use.epoch.dt.to_pydatetime())
        Tdata = conv.numpy_dt2dt(DForb_use.epoch.values)

        Xitrp = stats.lagrange_interpolate(Tdata, DForb_use["x"], titrp, n=n)
        Yitrp = stats.lagrange_interpolate(Tdata, DForb_use["y"], titrp, n=n)
        Zitrp = stats.lagrange_interpolate(Tdata, DForb_use["z"], titrp, n=n)

        Clkitrp = np.interp(
            conv.dt2posix(np.array(titrp)),
            conv.dt2posix(np.array(Tdata)),
            DForb_use["clk"].values,
        )

        # ClkDummy = np.array([999999.999999] * len(titrp))

        d = {"epoch": titrp, "x": Xitrp, "y": Yitrp, "z": Zitrp, "clk": Clkitrp}
        DForb_tmp = pd.DataFrame(d)

        ### sometihng else must be tested o give the annex val directly in the col of DForb_tmp
        DFannex_vals = DForb_use.drop(
            ["epoch", "x", "y", "z", "clk"], axis=1
        ).drop_duplicates()
        DFannex_vals = pd.concat(
            [DFannex_vals] * (len(titrp)), ignore_index=True, axis=0
        )

        DForb_tmp = pd.concat((DForb_tmp, DFannex_vals), axis=1)
        df_orb_stk.append(DForb_tmp)

        if plot:
            # plt.plot(Tdata,DForb_use.x,'o')
            # plt.plot(titrp,Xitrp,'.')
            ## GUS mod 220322
            fig, axr = plt.subplots(1, 1, sharex="all")
            Symb = axr.plot(Tdata, DForb_use.x, "o")
            Symb = axr.plot(titrp, Xitrp, ".")

    DForb_out = pd.concat(df_orb_stk)

    if append_to_input_df:
        DForb_out = pd.concat((orb_df_inp, DForb_out))

    DForb_out.reset_index(drop=True)
    DForb_out[["x", "y", "z", "clk"]] = DForb_out[["x", "y", "z", "clk"]].astype(float)
    return DForb_out


def orb_df_crf2trf(df_orb_inp, df_eop_inp, time_scale_inp="gps", inv_trf2crf=False):
    """
    Convert an Orbit DataFrame from Celetrial Reference Frame to
    Terrestrial Reference Frame.

    Requires EOP to work. Cf. note below.

    Parameters
    ----------
    df_orb_inp : DataFrame
        Input Orbit DataFrame in Celetrial Reference Frame.
    df_eop_inp : DataFrame
        EOP DataFrame  (C04 format).
    time_scale_inp : str, optional
        The time scale used in. manage 'utc', 'tai' and 'gps'.
        The default is "gps".
    inv_trf2crf : bool, optional
        Provide the inverse transformation TRF => CRF.
        The default is False.

    Returns
    -------
    DForb_out : DataFrame
        Output Orbit DataFrame in Terrestrial Reference Frame.
        (or Celestrial if inv_trf2crf is True)

    Note
    ----
    The EOP can be obtained from the IERS C04 products.
    e.g.
    https://datacenter.iers.org/data/latestVersion/224_EOP_C04_14.62-NOW.IAU2000A224.txt
    To get them as a Compatible DataFrame, use the function
    files_rw.read_eop_C04()
    """

    df_orb = df_orb_inp.copy()

    import geodezyx.reffram.sofa18 as sofa

    ### bring everything to UTC
    if time_scale_inp.lower() == "gps":
        df_orb["epoch_utc"] = conv.dt_gpstime2dt_utc(df_orb["epoch"])
    elif time_scale_inp.lower() == "tai":
        df_orb["epoch_utc"] = conv.dt_tai2dt_utc(df_orb["epoch"])
    elif time_scale_inp.lower() == "utc":
        df_orb["epoch_utc"] = df_orb["epoch"]
    ### TT and UT1 are not implemented (quite unlikely to have them as input)

    ### do the time scale's conversion
    df_orb["epoch_tai"] = conv.dt_utc2dt_tai(df_orb["epoch_utc"])
    df_orb["epoch_tt"] = conv.dt_tai2dt_tt(df_orb["epoch_tai"])
    df_orb["epoch_ut1"] = conv.dt_utc2dt_ut1_smart(df_orb["epoch_utc"], df_eop_inp)

    ### Do the EOP interpolation
    df_eop_intrp = eop_interpotate(df_eop_inp, df_orb["epoch_utc"])
    ### bring the EOP to radians
    Xeop = np.deg2rad(conv.arcsec2deg(df_eop_intrp["x"]))
    Yeop = np.deg2rad(conv.arcsec2deg(df_eop_intrp["y"]))

    TRFstk = []

    for tt, ut1, xeop, yeop, x, y, z in zip(
        df_orb["epoch_tt"],
        df_orb["epoch_ut1"],
        Xeop,
        Yeop,
        df_orb["x"],
        df_orb["y"],
        df_orb["z"],
    ):

        MatCRF22TRF = sofa.iau_c2t06a(
            2400000.5, conv.dt2mjd(tt), 2400000.5, conv.dt2mjd(ut1), xeop, yeop
        )
        if inv_trf2crf:
            MatCRF22TRF = np.linalg.inv(MatCRF22TRF)

        CRF = np.array([x, y, z])
        TRF = np.dot(MatCRF22TRF, CRF)

        TRFstk.append(TRF)

    ### Final stack and replacement
    TRFall = np.vstack(TRFstk)
    DForb_out = df_orb_inp.copy()
    DForb_out[["x", "y", "z"]] = TRFall

    return DForb_out


#### FCT DEF
def orb_df_reg_2_multidx(OrbDFin, index_order=["prn", "epoch"]):
    """
    From an regular Orbit DF generated by read_sp3(), set some columns
    (typically ["prn","epoch"]) as indexes
    The outputed DF is then a Multi-index DF
    """
    OrbDFwrk = OrbDFin.reset_index()
    OrbDFwrk = OrbDFwrk.sort_values(index_order)
    OrbDFwrk = OrbDFwrk.set_index(index_order, inplace=False)
    return OrbDFwrk


def orb_df_multidx_2_reg(OrbDFin, index_order=["prn", "epoch"]):
    """
    Convert a Multi-index formatted OrbitDF to his original form
    """
    OrbDFwrk = OrbDFin.reset_index()
    OrbDFwrk = OrbDFwrk.sort_values(index_order)
    OrbDFwrk["sys"] = OrbDFwrk["prn"].apply(lambda x: x[0])
    OrbDFwrk["prni"] = OrbDFwrk["prn"].apply(lambda x: int(x[1:]))
    return OrbDFwrk


def orb_df_common_epoch_finder(
    OrbDFa_in,
    orb_df_b_in,
    return_index=False,
    supplementary_sort=False,
    order=["prn", "epoch"],
    skip_reg2multidx_OrbDFa=False,
    skip_reg2multidx_OrbDFb=False,
):
    """
    This function finds common satellites and epochs in two Orbit DataFrames and outputs the corresponding Orbit DataFrames.

    Parameters
    ----------
    OrbDFa_in : DataFrame
        The first input Orbit DataFrame.
    orb_df_b_in : DataFrame
        The second input Orbit DataFrame.
    return_index : bool, optional
        If True, the function also returns the common index. Default is False.
    supplementary_sort : bool, optional
        If True, an additional sort is performed. This is useful for multi GNSS where the output DataFrame may not be well sorted. Default is False.
    order : list of str, optional
        The order of the index for the multi-index DataFrame. Default is ["prn","epoch"].
    skip_reg2multidx_OrbDFa : bool, optional
        If True, skips the conversion of the first input DataFrame to a multi-index DataFrame. Default is False.
        The inputs are assumed to be already in multi-index format to optimize execution speed.
        (For advanced use only)
    skip_reg2multidx_OrbDFb : bool, optional
        If True, skips the conversion of the second input DataFrame to a multi-index DataFrame. Default is False.
        The inputs are assumed to be already in multi-index format to optimize execution speed.
        (For advanced use only)

    Returns
    -------
    orb_df_a_out : DataFrame
        The first output Orbit DataFrame with common satellites and epochs.
    orb_df_b_out : DataFrame
        The second output Orbit DataFrame with common satellites and epochs.
    Iinter : Index, optional
        The common index. Only returned if return_index is True.

    Note
    ----
    designed for orbits/sp3 first with sat and epoch as order parmeter,
    but can be used also for instance for snx files with
    STAT and epoch as order parmeter
    """
    if not skip_reg2multidx_OrbDFa:
        orb_df_a = orb_df_reg_2_multidx(OrbDFa_in, index_order=order)
    else:
        orb_df_a = OrbDFa_in
    if not skip_reg2multidx_OrbDFb:
        orb_df_b = orb_df_reg_2_multidx(orb_df_b_in, index_order=order)
    else:
        orb_df_b = orb_df_b_in

    i1 = orb_df_a.index
    i2 = orb_df_b.index

    iinter = i1.intersection(i2)
    iinter = iinter.sort_values()

    orb_df_a_out = orb_df_a.loc[iinter]
    orb_df_b_out = orb_df_b.loc[iinter]

    if supplementary_sort:
        orb_df_a_out = orb_df_a_out.sort_values(order)
        orb_df_b_out = orb_df_b_out.sort_values(order)

    if len(orb_df_a_out) != len(orb_df_b_out):
        log.warning("len(Orb/ClkDFa_out) != len(Orb/ClkDFb_out)")
        log.warning("TIPS : ClkDFa_in and/or ClkDFb_in might contain duplicates")

    if return_index:
        return orb_df_a_out, orb_df_b_out, iinter
    else:
        return orb_df_a_out, orb_df_b_out


def orb_df_const_sv_columns_maker(orb_df_in, inplace=True):
    """
    (re)generate the const and sv columns from the sat one
    """
    if inplace:
        orb_df_in["sys"] = orb_df_in["prn"].str[0]
        orb_df_in["prni"] = orb_df_in["prn"].apply(lambda x: int(x[1:]))
        return None
    else:
        orb_df_out = orb_df_in.copy()
        orb_df_out["sys"] = orb_df_out["prn"].str[0]
        orb_df_out["prni"] = orb_df_out["prn"].apply(lambda x: int(x[1:]))
        return orb_df_out


#   _____ _            _      _____        _        ______
#  / ____| |          | |    |  __ \      | |      |  ____|
# | |    | | ___   ___| | __ | |  | | __ _| |_ __ _| |__ _ __ __ _ _ __ ___   ___  ___
# | |    | |/ _ \ / __| |/ / | |  | |/ _` | __/ _` |  __| '__/ _` | '_ ` _ \ / _ \/ __|
# | |____| | (_) | (__|   <  | |__| | (_| | || (_| | |  | | | (_| | | | | | |  __/\__ \
#  \_____|_|\___/ \___|_|\_\ |_____/ \__,_|\__\__,_|_|  |_|  \__,_|_| |_| |_|\___||___/

### Clock DataFrames


def clk_df_filter(
    clk_df_in,
    typ=("AS", "AR"),
    name=None,
    ac=None,
    epoch_strt=dt.datetime(1980, 1, 1),
    epoch_end=dt.datetime(2099, 1, 1),
    name_regex=False,
):
    """
    Filter the content of a Clock DataFrame

    Parameters
    ----------
    clk_df_in : DataFrame
        Input Clock DataFrame
        (a concatenation of DF generated by files_rw.read_clk.
    typ : iterable of str, optional
        List of the types of clocks: AS (satellite) or AR (receiver).
        The default is ("AS","AR").
    name : iterable of str, optional
        List of wished satellites/stations.
        Can be a regex (see also name_regex)
        The default is None.
    ac : iterable of str, optional
        List of wished ACs. The default is None.
    epoch_strt : datetime, optional
        Start epoch. The default is dt.datetime(1980,1,1).
    epoch_end : datetime, optional
        End epoch (not included). The default is dt.datetime(2099,1,1).
    name_regex : bool, optional
        the given names as 'name' arguments are regular expressions
        Some useful regex are given bellow
        The default is False

    Returns
    -------
    Clock DataFrame
        Output Clock DataFrame.

    Notes
    -----
    '^E[0-9]{2}': Galileo Satellites
    '^G[0-9]{2}': GPS Satellites

    """

    if type(clk_df_in) is str:
        clk_df_wrk = utils.pickle_loader(clk_df_in)
    else:
        clk_df_wrk = clk_df_in

    bool = np.ones(len(clk_df_wrk)).astype(bool)

    if typ:
        bool_tmp = clk_df_wrk.type.isin(typ)
        bool = bool & np.array(bool_tmp)

    if name:
        if not name_regex:  ### full name mode
            bool_tmp = clk_df_wrk.name.isin(name)
            bool = bool & np.array(bool_tmp)
        else:  ### REGEX mode
            bool_tmp = np.zeros(len(clk_df_wrk.name)).astype(bool)
            for rgx in name:
                nam_serie = clk_df_wrk.name
                bool_tmp = bool_tmp | np.array(nam_serie.str.contains(rgx))

            bool = bool & np.array(bool_tmp)

    if ac:
        bool_tmp = clk_df_wrk.ac.isin(ac)
        bool = bool & np.array(bool_tmp)

    ##epoch
    bool_tmp = (epoch_strt <= clk_df_wrk.epoch) & (clk_df_wrk.epoch < epoch_end)
    bool = bool & np.array(bool_tmp)

    return clk_df_wrk[bool]


def clk_df_filter2(
    clk_df_in,
    typ=("AS", "AR"),
    name=None,
    ac=None,
    epoch_strt=dt.datetime(1980, 1, 1),
    epoch_end=dt.datetime(2099, 1, 1),
    name_regex=False,
):
    """
    attempt for a faster version of clk_df_filter, but the original is faster
    """

    if type(clk_df_in) is str:
        clk_df_wrk = utils.pickle_loader(clk_df_in)
    else:
        clk_df_wrk = clk_df_in

    clkdf_stk = []

    for (ityp, iname, iac), clkdf_grp in clk_df_wrk.groupby(["type", "name", "ac"]):
        if typ:
            bool_typ = True if ityp in typ else False
        else:
            bool_typ = True

        if name:
            if not name_regex:
                bool_name = True if iname in name else False
            else:
                bool_name = any([re.search(n, iname) for n in name])
        else:
            bool_name = True

        if ac:
            bool_ac = True if iac in ac else False
        else:
            bool_ac = True

        if not (bool_typ and bool_name and bool_ac):
            continue
        else:
            if epoch_strt > dt.datetime(1980, 1, 1) or epoch_end < dt.datetime(
                2099, 1, 1
            ):
                bool_epoc = (epoch_strt <= clkdf_grp["epoch"]) & (
                    clkdf_grp["epoch"] < epoch_end
                )
                clkdf_stk.append(clkdf_grp[bool_epoc])
            else:
                clkdf_stk.append(clkdf_grp)

    clkdf_out = pd.concat(clkdf_stk)

    return clkdf_out


def clk_df_reg_2_multidx(clk_df_inp, index_order=["name", "epoch"]):
    """
    From an regular Clock DF generated by read_clk(), set some columns
    (typically ["name","epoch"]) as indexes
    The outputed DF is then a Multi-index DF

    It an adapted version of orb_df_reg_2_multidx
    """

    return orb_df_reg_2_multidx(clk_df_inp, index_order)


def clk_df_common_epoch_finder(
    ClkDFa_in,
    ClkDFb_in,
    return_index=False,
    supplementary_sort=False,
    order=["name", "epoch"],
):
    """
    Find common sats/station and epochs in to Clock DF, and output the
    corresponding Clock DFs

    Is an adapted version of orb_df_common_epoch_finder
    """

    return orb_df_common_epoch_finder(
        ClkDFa_in,
        ClkDFb_in,
        return_index=return_index,
        supplementary_sort=supplementary_sort,
        order=order,
    )


def clk_df_common_epoch_finder_multi(
    clk_df_list_in,
    return_index=False,
    supplementary_sort=False,
    order=["name", "epoch"],
):
    """
    Find common sats/station and epochs in to Clock DF, and output the
    corresponding Clock DFs

    Is is the multi version of clk_df_common_epoch_finder
    """

    clk_df_ref = clk_df_list_in[0]

    #### First loop: we find the common epochs
    for ClkDF in clk_df_list_in[1:]:

        OUTTUP = orb_df_common_epoch_finder(
            clk_df_ref,
            ClkDF,
            return_index=True,
            supplementary_sort=supplementary_sort,
            order=order,
        )

        clk_df_ref, _, Iinter = OUTTUP

    #### second loop: we use the common epochs found for the outputed ClkDF
    clk_df_list_out = []
    for ClkDF in clk_df_list_in:
        clk_df_out = ClkDF.set_index(order).loc[Iinter]
        clk_df_list_out.append(clk_df_out)

    if not return_index:
        return clk_df_list_out
    else:
        return clk_df_list_out, Iinter


#   _____ _      _____   __      __   _ _     _       _   _
#  / ____| |    |  __ \  \ \    / /  | (_)   | |     | | (_)
# | (___ | |    | |__) |  \ \  / /_ _| |_  __| | __ _| |_ _  ___  _ __
#  \___ \| |    |  _  /    \ \/ / _` | | |/ _` |/ _` | __| |/ _ \| '_ \
#  ____) | |____| | \ \     \  / (_| | | | (_| | (_| | |_| | (_) | | | |
# |_____/|______|_|  \_\     \/ \__,_|_|_|\__,_|\__,_|\__|_|\___/|_| |_|


def svn_prn_equiv_df(path_meta_snx):
    """
    generate a SVN <> PRN equivalent DataFrame

    Parameters
    ----------
    path_meta_snx : str
        path of the MGEX metadata sinex.
        last version avaiable here
        http://mgex.igs.org/IGS_MGEX_Metadata.php

    Returns
    -------
    DFfin : Pandas DataFrame
        SVN <> PRN equivalent DataFrame.

    """

    df_svn = files_rw.read_sinex_versatile(
        path_meta_snx, "SATELLITE/IDENTIFIER", header_line_idx=-2
    )

    df_prn = files_rw.read_sinex_versatile(
        path_meta_snx, "SATELLITE/PRN", header_line_idx=-2
    )

    df_svn.drop(columns="Comment__________________________________", inplace=True)
    df_prn.drop(columns="Comment_________________________________", inplace=True)

    ## the next lines 1 and 3 seems like they have became useless
    df_svn["SVN_"] = df_svn["SVN_"].apply(lambda x: x[0] + x[1:])
    df_prn.replace(dt.datetime(1970, 1, 1), dt.datetime(2099, 1, 1), inplace=True)
    df_prn["SVN_"] = df_prn["SVN_"].apply(lambda x: x[0] + x[1:])

    df_stk = []

    for isat, sat in df_prn.iterrows():
        svn = sat["SVN_"]

        sat["Block"] = df_svn[df_svn["SVN_"] == svn]["Block__________"].values[0]
        df_stk.append(sat)

    df_fin = pd.concat(df_stk, axis=1).transpose()

    df_fin.rename(
        columns={"SVN_": "SVN", "Valid_From____": "start", "Valid_To______": "end"},
        inplace=True,
    )

    df_fin["const"] = df_fin["SVN"].apply(lambda x: x[0])
    df_fin["SVN_int"] = df_fin["SVN"].apply(lambda x: int(x[1:]))
    df_fin["PRN_int"] = df_fin["PRN"].apply(lambda x: int(x[1:]))

    return df_fin


def svn_prn_equiv(sat_in, date_in, svn_prn_equiv_df, mode="svn2prn", full_output=False):
    """
    Get the equivalence SVN <> PRN for a given epoch

    Parameters
    ----------
    sat_in : str
        Satellite "ID", SVN or PRN.
    date_in : datetime
        wished epoch.
    svn_prn_equiv_df : DataFrame
        Equivalence table generated by svn_prn_equiv_df.
    mode : str, optional
        prn2svn: PRN > SVN
        svn2prn: SVN > PRN.
        The default is "svn2prn".
    full_output : bool, optional
        get the complete Equivalence table row. The default is False.

    Returns
    -------
    str or DataFrame
    """

    svnorprn1 = mode[:3].upper()
    svnorprn2 = mode[-3:].upper()

    df_sat = svn_prn_equiv_df[svn_prn_equiv_df[svnorprn1] == sat_in]
    bool_date = np.logical_and((df_sat.start <= date_in), (date_in < df_sat.end))
    df_out = df_sat[bool_date]

    if len(df_out) != 1:
        log.warning("several or no %s entries !!! %s %s", mode, sat_in, date_in)

    if full_output:
        return df_out
    else:
        return df_out[svnorprn2].values[0]


def get_block_svn(sat_in, svn_prn_equiv_df):
    """
    Get the equivalence SVN block type

    Parameters
    ----------
    sat_in : str
        Satellite SVN.
    svn_prn_equiv_df : DataFrame
        Equivalence table generated by svn_prn_equiv_df.
    Returns
    -------
    str with the block name
    """
    df_sat = svn_prn_equiv_df[svn_prn_equiv_df["SVN"] == sat_in]
    # if df_sat.empty:
    #     print('SVN NOT FOUND')
    # else:
    block = df_sat.Block.values[0]

    return block


def stats_slr(df_in, grpby_keys=["sat"], threshold=0.5):
    """
    computes statistics for SLR Residuals

    Parameters
    ----------
    df_in : Pandas DataFrame
        Input residual Dataframe from read_pdm_res_slr.
    grpby_keys : list of str, optional
        The default is ['sat'].
        per day, per solution, per satellite: ['day','sol','sat']
        per day, per solution, per station: ['day','sol','sta']
        per day, per solution, per satellite, per station: ['day','sol','sta','sat']
    threshold : float
        apply a Threshold

    Returns
    -------
    dd : Output statistics DataFrame
        return the mean, the rms and the std.
    """

    dd = df_in[np.abs(df_in["res"]) < threshold]

    dd_grp = dd.groupby(grpby_keys)
    dd_mean = dd_grp["res"].agg(np.mean).rename("mean") * 1000
    dd_rms = dd_grp["res"].agg(stats.rms_mean).rename("rms") * 1000
    dd_std = dd_grp["res"].agg(np.std).rename("std") * 1000
    dd = pd.concat([dd_mean, dd_std, dd_rms], axis=1)
    dd.reset_index(inplace=True)

    return dd


#  ______           _   _        ____       _            _        _   _               _____                               _
# |  ____|         | | | |      / __ \     (_)          | |      | | (_)             |  __ \                             | |
# | |__   __ _ _ __| |_| |__   | |  | |_ __ _  ___ _ __ | |_ __ _| |_ _  ___  _ __   | |__) |_ _ _ __ __ _ _ __ ___   ___| |_ ___ _ __ ___
# |  __| / _` | '__| __| '_ \  | |  | | '__| |/ _ \ '_ \| __/ _` | __| |/ _ \| '_ \  |  ___/ _` | '__/ _` | '_ ` _ \ / _ \ __/ _ \ '__/ __|
# | |___| (_| | |  | |_| | | | | |__| | |  | |  __/ | | | || (_| | |_| | (_) | | | | | |  | (_| | | | (_| | | | | | |  __/ ||  __/ |  \__ \
# |______\__,_|_|   \__|_| |_|  \____/|_|  |_|\___|_| |_|\__\__,_|\__|_|\___/|_| |_| |_|   \__,_|_|  \__,_|_| |_| |_|\___|\__\___|_|  |___/


### EOP / Earth Oreintation Parameters


def eop_interpotate(df_eop, epochs_intrp, eop_params=["x", "y"]):
    """
    Interopolate the EOP provided in a C04-like DataFrame

    Parameters
    ----------
    df_eop : DataFrame
        Input EOP DataFrame (C04 format).
        Can be generated by files_rw.read_eop_C04
    epochs_intrp : datetime of list of datetimes
        Wished epochs for the interpolation.
    eop_params : list of str, optional
        Wished EOP parameter to be interpolated.
        The default is ["x","y"].

    Returns
    -------
    out : DataFrame or Series
        Interpolated parameters.
        Series if onely one epoch is provided, df_eop elsewere
    """

    from geodezyx import interp

    if not utils.is_iterable(epochs_intrp):
        singleton = True
    else:
        singleton = False

    i_eop = dict()
    out_eop = dict()
    out_eop["epoch"] = epochs_intrp

    for eoppar in eop_params:
        intrp = interp.Interp1dTime(df_eop.epoch, df_eop[eoppar])
        i_eop[eoppar] = intrp
        try:
            out_eop[eoppar] = intrp(epochs_intrp)
        except ValueError as err:
            log.error("in EOP interpolation")
            log.error("param.: %s, epoch: %s", eoppar, epochs_intrp)
            raise err

    if not singleton:
        out = pd.DataFrame(out_eop)
    else:
        out = pd.Series(out_eop)

    return out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 25 20:37:05 2023

@author: psakic

This sub-module of geodezyx.conv deals with UTM projections.

it can be imported directly with:
from geodezyx import conv

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

#### Import the logger
import logging

########## BEGIN IMPORT ##########
#### External modules
import numpy as np

#### geodeZYX modules
# from geodezyx import utils,conv
log = logging.getLogger('geodezyx')

##########  END IMPORT  ##########


def utm_geo2xy(lat,lon,ellips="wgs84",zone=None):
    """
    Latitude/Longitude to Universal Transverse Mercator(UTM)
    coordinates precise (mm-level) conversion.

    Converts coordinates in Latitude/Longitude (in degrees)
    to UTM X and Y (in meters).

    Default ellipsoid is WGS84.
    
    Parameters
    ----------
    lat : float or numpy array
        Latitude (in degrees).
    lon : float or numpy array
        Longitude (in degrees).
    ellips : str or 2-tuple, optional
        uses specific ellipsoid for conversion. ellipsoid can be one
    	of the following strings
		'wgs84': World Geodetic System 1984 (default)
		'nad27': North American ellips 1927
		'clk66': Clarke 1866
		'nad83': North American ellips 1983
		'grs80': Geodetic Reference System 1980
		'int24': International 1924 / Hayford 1909
        OR
        can be a 2-element tuple [A,F] where A is semimajor axis (in meters)
        F is flattening of the user-defined ellipsoid.
        The default is "wgs84".
    zone : int, optional
        forces the UTM ZONE (scalar integer or same size as
        LAT and LON) instead of automatic set. 
        Positive for the nothern hemisphere.
        Negative for the southern hemisphere.
        The default is None.

    Returns
    -------
    x,y : float or numpy array
        UTM X and Y coordinates (in meters).
    f : float or numpy array
        returns also the computed UTM ZONE (negative
        value for southern hemisphere points).
        
    Notes
    -----
    It does not perform cross-ellips conversion.
    Precision is near a millimeter.
    
    This function implements the IGN algorithm.
    This one is *more accurate* than `pyproj` or `utm` python lib 
    because it implements more coefficients.
    
    Source
    ------
    Adapted from WebObs source code, François Beauducel et al. 
    https://github.com/IPGP/webobs/blob/master/CODE/matlab/ll2utm.m
    
	I.G.N., Projection cartographique Mercator Transverse: Algorithmes,
    Notes Techniques NT/G 76, janvier 1995.    
    https://geodesie.ign.fr/contenu/fichiers/documentation/algorithmes/notice/NTG_76.pdf
    
    """

    # constants
    D0 = 180 / np.pi
    K0 = 0.9996
    X0 = 500000

    # defaults
    zone = []

    # ellipsoid parameters
    A1,F1 = ellipsoid_params(ellips)
    
    p1 = lat / D0
    l1 = lon / D0
    
    # UTM zone automatic setting
    if not zone:
        F0 = np.round((l1 * D0 + 183) / 6)
    else:
        F0 = zone
    
    B1 = A1 * (1 - 1 / F1)
    E1 = np.sqrt((A1 * A1 - B1 * B1) / (A1 * A1))
    P0 = 0 / D0
    L0 = (6 * F0 - 183) / D0
    
    Y0 = 10000000.0 * (p1 < 0)
    
    N = K0 * A1
    C = utm_coef(E1,0)
    B = C[0] * P0 + C[1] * np.sin(2 * P0) + C[2] * np.sin(4 * P0) + C[3] * np.sin(6 * P0) + C[4] * np.sin(8 * P0)
    YS = Y0 - N * B
    C = utm_coef(E1,2)
    L = np.log(np.multiply(np.tan(np.pi / 4 + p1 / 2),(((1 - E1 * np.sin(p1)) / (1 + E1 * np.sin(p1))) ** (E1 / 2))))
    z = complex(np.arctan(np.sinh(L) / np.cos(l1 - L0)),np.log(np.tan(np.pi / 4 + np.arcsin(np.sin(l1 - L0) / np.cosh(L)) / 2)))
    Z = np.multiply(np.multiply(N,C[0]),z) + np.multiply(N,(C[1] * np.sin(2 * z) + C[2] * np.sin(4 * z) + C[3] * np.sin(6 * z) + C[4] * np.sin(8 * z)))
    xs = np.imag(Z) + X0
    ys = np.real(Z) + YS
    
    f = np.multiply(F0,np.sign(lat))
    fu = np.unique(f)
    if np.isscalar(fu):
        f = fu
    
    x = xs
    y = ys
        
    return x,y,f


def utm_xy2geo(x, y, f, ellips='wgs84'):
    """
    Universal Transverse Mercator (UTM) coordinates to Latitude/Longitude

    Converts coordinates in UTM X and Y (in meters)
    to Latitude/Longitude (in degrees)

    Default ellipsoid is WGS84.

    Parameters
    ----------
    x : array-like
        UTM X coordinates (in meters).
    y : array-like
        UTM Y coordinates (in meters).
    f : int
        UTM zone.
        Positive for the nothern hemisphere.
        Negative for the southern hemisphere.
    ellips : str or tuple, optional
        Datum for conversion. Default is 'wgs84'. Can be one of the predefined strings or a tuple (A, F).

    Returns
    -------
    lat : array-like
        Latitude coordinates (in degrees).
    lon : array-like
        Longitude coordinates (in degrees).

    Source
    ------
    Adapted from WebObs source code, François Beauducel et al.
    https://github.com/IPGP/webobs/blob/master/CODE/matlab/ll2utm.m

	I.G.N., Projection cartographique Mercator Transverse: Algorithmes,
    Notes Techniques NT/G 76, janvier 1995.
    https://geodesie.ign.fr/contenu/fichiers/documentation/algorithmes/notice/NTG_76.pdf
    """

    A1,F1 = ellipsoid_params(ellips)

    D0 = 180 / np.pi
    maxiter = 100
    eps = 1e-11

    K0 = 0.9996
    X0 = 500000
    Y0 = 1e7 * (f < 0)
    P0 = 0
    L0 = (6 * abs(f) - 183) / D0
    E1 = np.sqrt((A1**2 - (A1 * (1 - 1 / F1))**2) / A1**2)
    N = K0 * A1

    C = utm_coef(E1, 0)
    YS = Y0 - N * (C[0] * P0 + C[1] * np.sin(2 * P0) + C[2] * np.sin(4 * P0) + C[3] * np.sin(6 * P0) + C[4] * np.sin(8 * P0))

    C = utm_coef(E1, 1)
    zt = (y - YS) / N / C[0] + 1j * (x - X0) / N / C[0]
    z = zt - C[1] * np.sin(2 * zt) - C[2] * np.sin(4 * zt) - C[3] * np.sin(6 * zt) - C[4] * np.sin(8 * zt)
    L = np.real(z)
    LS = np.imag(z)

    l = L0 + np.arctan(np.sinh(LS) / np.cos(L))
    p = np.arcsin(np.sin(L) / np.cosh(LS))

    L = np.log(np.tan(np.pi / 4 + p / 2))

    p = 2 * np.arctan(np.exp(L)) - np.pi / 2
    p0 = np.nan
    n = 0
    
    while np.any(np.logical_or(np.isnan(p0) , np.abs(p - p0) > eps)) and n < maxiter:
        p0 = p
        es = E1 * np.sin(p0)
        p = 2 * np.arctan(((1 + es) / (1 - es))**(E1 / 2) * np.exp(L)) - np.pi / 2
        n += 1

    lat = p * D0
    lon = l * D0

    return lat, lon

    ###########################################################################


def ellipsoid_params(ellips):
    # Available ellipsoids
    ellips_dic = dict()
    ellips_dic['wgs84'] = np.array([6378137.0, 298.257223563])
    ellips_dic['nad83'] = np.array([6378137.0, 298.257222101])
    ellips_dic['grs80'] = np.array([6378137.0, 298.257222101])
    ellips_dic['nad27'] = np.array([6378206.4, 294.978698214])
    ellips_dic['int24'] = np.array([6378388.0, 297.0])
    ellips_dic['clk66'] = np.array([6378206.4, 294.978698214])

    if type(ellips) is str:
        # LL2UTM(...,ellips) with ellips as char
        if not ellips in ellips_dic.keys():
            raise Exception('Unkown ellips name "%s"', ellips)
        A1 = ellips_dic[ellips][0]
        F1 = ellips_dic[ellips][1]
    else:
        # LL2UTM(...,ellips) with ellips as [A,F] user-defined
        A1 = ellips[1]
        F1 = ellips[2]

    return A1, F1


def utm_coef(e=None, m=0):
    """
    Projection coefficients
	returns a vector of 5 coefficients based on the input parameters

    Parameters
    ----------
    e : TYPE, optional
        first ellipsoid excentricity. The default is None.
    m : int, optional
        Describes the expected coefficients.
		M = 0 for transverse mercator
		M = 1 for transverse mercator reverse coefficients
		M = 2 for merdian arc
        The default is 0.

    Notes
    -----
    It does not perform cross-ellips conversion.
    Precision is near a millimeter.

    This function implements the IGN algorithm.
    This one is *more accurate* than `pyproj` or `utm` python lib
    because it implements more coefficients.

    Returns
    -------
    c : numpy array
        Coefficient Matrix.

    """

    if 0 == m:
        c0 = np.array([[- 175 / 16384, 0, - 5 / 256, 0, - 3 / 64, 0, - 1 / 4, 0, 1],
                       [- 105 / 4096, 0, - 45 / 1024, 0, - 3 / 32, 0, - 3 / 8, 0, 0],
                       [525 / 16384, 0, 45 / 1024, 0, 15 / 256, 0, 0, 0, 0],
                       [- 175 / 12288, 0, - 35 / 3072, 0, 0, 0, 0, 0, 0],
                       [315 / 131072, 0, 0, 0, 0, 0, 0, 0, 0]])
    elif 1 == m:
        c0 = np.array([[- 175 / 16384, 0, - 5 / 256, 0, - 3 / 64, 0, - 1 / 4, 0, 1],
                       [1 / 61440, 0, 7 / 2048, 0, 1 / 48, 0, 1 / 8, 0, 0],
                       [559 / 368640, 0, 3 / 1280, 0, 1 / 768, 0, 0, 0, 0],
                       [283 / 430080, 0, 17 / 30720, 0, 0, 0, 0, 0, 0],
                       [4397 / 41287680, 0, 0, 0, 0, 0, 0, 0, 0]])
    elif 2 == m:
        c0 = np.array([[- 175 / 16384, 0, - 5 / 256, 0, - 3 / 64, 0, - 1 / 4, 0, 1],
                       [- 901 / 184320, 0, - 9 / 1024, 0, - 1 / 96, 0, 1 / 8, 0, 0],
                       [- 311 / 737280, 0, 17 / 5120, 0, 13 / 768, 0, 0, 0, 0],
                       [899 / 430080, 0, 61 / 15360, 0, 0, 0, 0, 0, 0],
                       [49561 / 41287680, 0, 0, 0, 0, 0, 0, 0, 0]])
    else:
        log.error('m must be 0, 1 or 2')
        raise ValueError('m must be 0, 1 or 2')

    c = np.polyval(c0.T, e)

    return c

xyf = utm_geo2xy( 16.043829, -61.663406)
utm_xy2geo(*xyf)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.files_rw contains functions to 
read RINEX files observation files.

it can be imported directly with:
from geodezyx import files_rw

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

#### Import the logger
import logging
import os
import pathlib
import re
from io import StringIO

import hatanaka

########## BEGIN IMPORT ##########
#### External modules
import numpy as np
import pandas as pd
from tqdm import tqdm

#### geodeZYX modules
from geodezyx import operational, utils, conv

log = logging.getLogger('geodezyx')


def read_rinex_obs(rnx_in, set_index=None , return_header = False):
    """
    Frontend function to read a RINEX Observation, version 2 or 3

    Parameters
    ----------
    rnx_in : see below
        input RINEX.
        can be the path of a RINEX file as string or as Path object,
        or directly the RINEX content as a string, bytes, StringIO object or a
        list of lines
    set_index : str or list of str, optional
        define the columns for the index.
        If None, the output DataFrame is "flat", with integer index
        ["epoch","prn"] for instance set the epoch and the prn as MultiIndex
        The default is None.
    return_header : bool, optional
        If True, the function returns a tuple (df_rnx_obs, header_lines),
        where header_lines is a list of strings containing the header lines.

    Returns
    -------
    df_rnx_obs : Pandas DataFrame / GeodeZYX's RINEX format
    header_lines : list of str, optional
        If return_header is True, this contains the header lines of the RINEX file.
        Otherwise, it is not returned.
    """

    if conv.rinex_regex_search_tester(rnx_in,short_name=True,long_name=False):
        return read_rinex2_obs(rnx_in, set_index=set_index, return_header=return_header)
    elif conv.rinex_regex_search_tester(rnx_in,short_name=False,long_name=True):
        return read_rinex3_obs(rnx_in, set_index=set_index, return_header=return_header)
    else:
        log.error("RINEX version not recognized: %s", rnx_in)
        log.error("use read_rinex2_obs or read_rinex3_obs functions manually")
        return None


def read_rinex2_obs(rnx_in, set_index=None, return_header = False):
    """
    Read a RINEX Observation, version 2

    Parameters
    ----------
    rnx_in : see below
        input RINEX.
        can be the path of a RINEX file as string or as Path object,
        or directly the RINEX content as a string, bytes, StringIO object or a
        list of lines
    set_index : str or list of str, optional
        define the columns for the index.
        If None, the output DataFrame is "flat", with integer index
        ["epoch","prn"] for instance set the epoch and the prn as MultiIndex
        The default is None.
    return_header : bool, optional
        If True, the function returns a tuple (df_rnx_obs, header_lines),
        where header_lines is a list of strings containing the header lines.

    Returns
    -------
    df_rnx_obs : Pandas DataFrame / GeodeZYX's RINEX format
    header_lines : list of str, optional
        If return_header is True, this contains the header lines of the RINEX file.
        Otherwise, it is not returned.
    """

    #### open rinex
    lines, epochs_dt, filename = open_rinex(rnx_in)

    ### get the header
    l_head = _header_reader(lines)

    # get the rinex type (mono or multi GNSS)
    rnx_typ = l_head[0][40]

    mixed_rnx = rnx_typ == "M"
    # True if M, False if not

    #### get the observables
    l_obs = [l for l in l_head if "# / TYPES OF OBSERV" in l]
    ## clean SYS / # / OBS TYPES
    l_obs = [l[:60] for l in l_obs]

    obs_all_list_raw = " ".join(l_obs).split()
    obs_all_list = obs_all_list_raw[1:]
    obs_all_list = [
        e
        for sublist in [(e, e + "_LLI", e + "_SSI") for e in obs_all_list]
        for e in sublist
    ]

    nobs = int(obs_all_list_raw[0])
    nl_obs = int(np.ceil(nobs / 5))
    ## 5 is the max num of obs in the RINEX specs
    col_width = nobs * [14, 1, 1]

    df_all_stk = []

    #### reading the epochs
    for iepoc in tqdm(range(len(epochs_dt)), desc="Reading " + filename):
        epoch = epochs_dt[iepoc, 0]
        ## define the start/end indices of the epoch block
        il_srt = epochs_dt[iepoc, 1]
        if iepoc == len(epochs_dt) - 1:
            il_end = None
        else:
            il_end = epochs_dt[iepoc + 1, 1]

        ### get the lines of the epoch block
        lines_epoc = lines[il_srt:il_end]

        ### get the satellites for this epoch block
        sats_find_out = _sats_find(lines_epoc)
        flag, epoc, nsat, _, sats_split, iline_sats_end = sats_find_out        

        if flag != 0: # skip the epoch (which is not an observable one)
            log.warning("epoch %s skipped, flag %i",epoc, flag)
            continue

        ### for each sat, merge the breaked lines
        resub80 = lambda x: re.sub(r"[\r\n]+", "", x.ljust(80))
        # .replace("\r", "").replace("\n", "") i.e. re.sub(r"[\r\n]+", "", s) IS NOT .strip() !!
        # .strip() removes the spaces at the beginning and end of the string, and they should be kept!
        l_obs = [ resub80(e) for e in lines_epoc[iline_sats_end + 1: il_end]]
        l_obs_merg = ["".join(l_obs[n * nl_obs:(n + 1) * nl_obs]) for n in range(nsat)]

        ## read the epoch block using pandas' fixed width reader
        b = StringIO("\n".join(l_obs_merg))
        df_epoch = pd.read_fwf(b, header=None, widths=col_width)
        df_epoch.columns = obs_all_list

        # slow
        # df_epoch["prn"] = sats_split
        # df_epoch["epoch"] = epoch
        # faster
        col_epoch = pd.Series([epoch] * len(df_epoch), name="epoch")
        col_prn = pd.Series(sats_split, name="prn")

        if np.any(col_prn.str.len() == 0):
            log.critical("bad epoch %s" , epoc)
            raise Exception()

        df_epoch = pd.concat([col_epoch, col_prn, df_epoch], axis=1)

        df_all_stk.append(df_epoch)

    ## final concat and cosmetic (reorder columns, sort)
    df_rnx_obs = pd.concat(df_all_stk)

    # strip to have just 2 or 3 prn characters
    df_rnx_obs["prn"] = df_rnx_obs["prn"].str.strip()

    # mixed, system is in the epoch line
    if mixed_rnx or df_rnx_obs["prn"].str.len().max() == 3:
        df_rnx_obs["sys"] = df_rnx_obs["prn"].str[0]
    else:  # not mixed, system is the header RINEX type
        df_rnx_obs["sys"] = rnx_typ
        df_rnx_obs["prn"] = rnx_typ + df_rnx_obs["prn"].str.strip()

    df_rnx_obs["prni"] = df_rnx_obs["prn"].str[-2:].astype(int)

    main_cols = ["epoch", "sys", "prn", "prni"]
    df_rnx_obs = df_rnx_obs.reindex(main_cols + list(sorted(obs_all_list)), axis=1)
    df_rnx_obs.sort_values(["epoch", "prn"], inplace=True)
    df_rnx_obs.reset_index(drop=True, inplace=True)

    if set_index:
        df_rnx_obs.set_index(set_index, inplace=True)
        df_rnx_obs.sort_index(inplace=True)

    if return_header:
        return df_rnx_obs, l_head
    else:
        return df_rnx_obs

def read_rinex3_obs(rnx_in, set_index=None, return_header = False):
    """

    Read a RINEX Observation, version 3 or 4

    Parameters
    ----------
    rnx_in : see below
        input RINEX.
        can be the path of a RINEX file as string or as Path object,
        or directly the RINEX content as a string, bytes, StringIO object or a
        list of lines
    set_index : str or list of str, optional
        define the columns for the index.
        If None, the output DataFrame is "flat", with integer index
        ["epoch","prn"] for instance set the epoch and the prn as MultiIndex
        The default is None.
    return_header : bool, optional
        If True, the function returns a tuple (df_rnx_obs, header_lines),
        where header_lines is a list of strings containing the header lines.

    Returns
    -------
    df_rnx_obs : Pandas DataFrame / GeodeZYX's RINEX format
    header_lines : list of str, optional
        If return_header is True, this contains the header lines of the RINEX file.
        Otherwise, it is not returned.
    """

    #### open rinex
    lines, epochs, filename = open_rinex(rnx_in)

    ### get the header
    l_head = _header_reader(lines)

    #### get the systems and observations
    l_sys = [l for l in l_head if "SYS / # / OBS TYPES" in l]
    ## clean SYS / # / OBS TYPES
    l_sys = [l[:60] for l in l_sys]

    ## manage the 2 lines systems
    for il, l in enumerate(l_sys):
        if l[0] == " ":
            l_sys[il - 1] = l_sys[il - 1] + l
            l_sys.remove(l)

    #### store system and observables in a dictionnary
    dict_sys_nobs = dict()
    dict_sys_cols = dict()  # adds the prn, and LLI and SSI indicators
    obs_all_list = []

    for il, l in enumerate(l_sys):
        sysobs = l.split()
        # sysobs[0] = system letter
        # sysobs[1] = observables number
        # sysobs[2:] = observables names

        dict_sys_nobs[sysobs[0]] = int(sysobs[1])
        ## adds the LLI and SSI indicators
        obs_w_lli_ssi = [(e, e + "_LLI", e + "_SSI") for e in sysobs[2:]]
        obs_all_list.extend([e for sublist in obs_w_lli_ssi for e in sublist])

        dict_sys_cols_tmp = [("sys","epoch","prn",)] + obs_w_lli_ssi
        dict_sys_cols[sysobs[0]] = [e for sublist in dict_sys_cols_tmp for e in sublist]

        if len(sysobs[2:]) != int(sysobs[1]):
            log.warning(
                "difference between theorectical and actual obs nbr in the header"
            )

    ## the max number of observable (for the reading)
    nobs_max = max(dict_sys_nobs.values())

    df_all_stk = []
    error_sys_in_obs = False

    #### reading the epochs
    for iepoc in tqdm(range(len(epochs)), desc="Reading " + filename):
        epoch = epochs[iepoc, 0]
        
        ## check the flag
        flag = int(lines[epochs[iepoc, 1]][31])
        if flag != 0: # skip the epoch (which is not an observable one)
            log.warning("epoch %s skipped, flag %i", epoch, flag)
            continue

        ## define the start/end indices of the epoch block
        iline_start = epochs[iepoc, 1] + 1
        if iepoc == len(epochs) - 1:
            iline_end = None
        else:
            iline_end = epochs[iepoc + 1, 1]

        l_epoc = [re.sub(r"[\r\n]+", "", e) for e in lines[iline_start:iline_end]]
        # .replace("\r", "").replace("\n", "") i.e. re.sub(r"[\r\n]+", "", s) IS NOT .strip() !!
        # .strip() removes the spaces at the beginning and end of the string, and they should be kept!

        for line in l_epoc:
            if "SYS / # / OBS TYPES" in line:
                error_sys_in_obs = True
                break

        ## read the epoch block using pandas' fixed width reader
        b = StringIO("\n".join(l_epoc))

        columns_width = [3] + nobs_max * [14, 1, 1]
        df_epoch = pd.read_fwf(b, header=None, widths=columns_width)

        col_epoch = pd.Series([epoch] * len(df_epoch), name="epoch")
        df_epoch_ok = pd.concat([col_epoch, df_epoch], axis=1)

        df_all_stk.append(df_epoch_ok)

    if error_sys_in_obs:
        log.error(
            "SYS / # / OBS TYPES in body, observable's order is modified, read_rinex does not manage this yet")
        return None

    ## final concat and cosmetic (reorder columns, sort)
    df_rnx_obs = pd.concat(df_all_stk)

    ## extract the system
    col_sys = pd.Series(df_rnx_obs[0].str.strip().str[0], name="sys")
    df_rnx_obs = pd.concat([col_sys, df_rnx_obs], axis=1)

    #### assign the correct observable names for each system
    df_rnx_obs_ok_stk = []
    for sys, df_rnx_obs_sys in df_rnx_obs.groupby("sys"):
        df_rnx_obs_sys_clean = df_rnx_obs_sys.iloc[:, : len(dict_sys_cols[sys])]
        df_rnx_obs_sys_clean.columns = dict_sys_cols[sys] # sys, epoch & prn are already in dict_sys_cols
        df_rnx_obs_ok_stk.append(df_rnx_obs_sys_clean)

    df_rnx_obs_ok = pd.concat(df_rnx_obs_ok_stk)

    col_prni = pd.Series(df_rnx_obs_ok["prn"].str.strip().str[-2:], name="prni", dtype=int)
    df_rnx_obs_ok = pd.concat([df_rnx_obs_ok, col_prni], axis=1)

    main_cols = ["epoch", "sys", "prn", "prni"]
    obs_used_list = list(set(df_rnx_obs_ok.columns).difference(set(main_cols)))
    df_rnx_obs_ok = df_rnx_obs_ok.reindex(main_cols + list(sorted(obs_used_list)), axis=1)

    df_rnx_obs_ok.sort_values(["epoch", "prn"], inplace=True)
    df_rnx_obs_ok.reset_index(drop=True, inplace=True)

    if set_index:
        df_rnx_obs_ok.set_index(set_index, inplace=True)
        df_rnx_obs_ok.sort_index(inplace=True)


    if return_header:
        return df_rnx_obs_ok, l_head
    else:
        return df_rnx_obs_ok


############ UTILITY FUNCTIONS


def open_rinex(rnx_inp,verbose=False):
    #### open block
    try:
        rnx_wrk = hatanaka.decompress(rnx_inp)
    except:
        rnx_wrk = rnx_inp
        pass

    lines = utils.open_readlines_smart(rnx_wrk, verbose=verbose)
    epochs_dt = operational.rinex_read_epoch(rnx_wrk, out_index=True)
    if type(rnx_inp) is str or type(rnx_inp) is pathlib.Path:
        filename = os.path.basename(rnx_inp)
    else:
        filename = "unknown filename"

    return lines, epochs_dt, filename


def df_rnx_clean_lli_ssi(df_rnx_in):
    """
    Remove the Loss of Lock Indicator (LLI) and Signal Strength Indicator (SSI)
    columns in a DataFrame RINEX

    Parameters
    ----------
    df_rnx_in : Pandas DataFrame / GeodeZYX's RINEX format
        A RINEX DataFrame with LLI/SSI columns.

    Returns
    -------
    df_rnx_out : Pandas DataFrame / GeodeZYX's RINEX format
        A RINEX DataFrame without LLI/SSI columns.
    """
    cols = df_rnx_in.columns
    cols_clean = [e for e in cols if not "LLI" in e and not "SSI" in e]
    df_rnx_out = df_rnx_in[cols_clean]
    return df_rnx_out


def observables_dict_per_sys(df_rnx_in):
    """
    Gives the GNSS observables for each GNSS system in a dictionnary


    Parameters
    ----------
    df_rnx_in : Pandas DataFrame / GeodeZYX's RINEX format
        A RINEX DataFrame.

    Returns
    -------
    dict_sys_obs : dict
        A dictionnary with GNSS system as key (G,R,E...).
        And the observalbes for each system as values

    Note
    ----

    Use dict_sys_obs_clean_LLI_SSI if your want to remove the LLI & SSI values
    """

    dict_sys_obs = dict()

    for sys in df_rnx_in["sys"].unique():
        df_sys = df_rnx_in[df_rnx_in["sys"] == sys]
        df_sys_mini = df_rnx_clean_lli_ssi(df_sys)

        obs_sys0 = pd.Series(df_sys_mini.isna().sum() == len(df_sys)).apply(np.logical_not)
        obs_sys = obs_sys0.index[obs_sys0][
            3:
        ]  ### strating from 3 to clean epoch sys prn

        # init_tup = ("epoch","sys","prn")
        # init_tup = []
        # obs_sys_full = [init_tup] + [(e,e+"_LLI",e+"_SSI") for e in obs_sys]
        obs_sys_full = [(e, e + "_LLI", e + "_SSI") for e in obs_sys]
        obs_sys_full = [e for sublist in obs_sys_full for e in sublist]

        dict_sys_obs[sys] = obs_sys_full

    return dict_sys_obs


def dict_sys_obs_clean_lli_ssi(dict_sys_obs_in):
    """
    Clean a `dict_sys_obs` (generated by `observables_dict_per_sys`)
    of its LLI and SSI values

    Parameters
    ----------
    dict_sys_obs_in : dict
        A dictionnary with GNSS system as key (G,R,E...).
        And the observalbes for each system as values.

    Returns
    -------
    dict_sys_obs_out : dict
        Same dictionnary cleanned of its LLI and SSI values.

    """
    dict_sys_obs_out = dict()

    for sys, obs in dict_sys_obs_in.items():
        obs_clean = [e for e in obs if not "LLI" in e and not "SSI" in e]
        dict_sys_obs_out[sys] = obs_clean

    return dict_sys_obs_out


############ INTERNAL FUNCTIONS


def _sats_find(lines_inp):
    """
    For RINEX2 only
    search for the satellites for each epoch block by reading the
    EPOCH/SAT record

    Parameters
    ----------
    lines_inp : List of str
        the lines of one epoch block (EPOCH/SAT + OBSERVATION records).

    Returns
    -------
    bloc_tuple : tuple
        a 5-tuple containing:
            epoc : datetime, the epoch of block
            nsat : int, the number of satellites
            lineconcat : str, the satellites as a concatenated string
            sats_split : list of str, the satellites as a list
            il : int, the index of the last line of the EPOCH/SAT record

    """
    iline_bloc = 0
    nlines_bloc = -1

    re_epoch = "^ {1,2}([0-9]{1,2} * ){5}"

    ### dummy initialisation of variables
    line_bloc = []
    lineconcat_stk = []
    epoc = None
    nsat = 0

    for il, l in enumerate(lines_inp):
        #####
        # re_sat="[A-Z][0-9][0-9]"
        # bool_epoch = re.search(re_epoch, l)
        # bool_sat=re.search(re_sat,l)

        ### we found an epoch line
        if re.search(re_epoch, l):
            flag = int(l[28])
            nsat = int(l[30:32])
            iline_bloc = 0
            nlines_bloc = int(np.ceil(nsat / 12))
            line_bloc = []
            #lineconcat = ""
            lineconcat_stk = []
            epoc = operational.read_rnx_epoch_line(l, rnx2=True)

        ### we read the sat lines based on the number of sat
        if iline_bloc <= nlines_bloc:
            line_bloc.append(l[32:].strip())
            # lineconcat = lineconcat + l[32:].strip()
            lineconcat_stk.append(l[32:].strip())
            iline_bloc += 1

        ### we stack everything when the sat block is over
        if iline_bloc == nlines_bloc:
            #in_epoch = False
            lineconcat = "".join(lineconcat_stk)
            sats_split = [lineconcat[3 * n : 3 * n + 3] for n in range(nsat)]
            bloc_tuple = (flag, epoc, nsat, lineconcat, sats_split, il)
            return bloc_tuple


def _header_reader(lines_inp):
    #### Split header and Observation body
    i_end_header = 0
    for il, l in enumerate(lines_inp):
        if "END OF HEADER" in l:
            i_end_header = il
            break
    return lines_inp[: i_end_header + 1]


### FUNCTION GRAVEYARD


def _line_reader(linein, nobs):
    """
    DISCONTINUED

    For RINEX3 Only

    read the content of an observation line. pd.read_fwf does the job
    """
    columns_width = [3] + nobs * [14, 1, 1]
    columns_cumul = [0] + list(np.cumsum(columns_width))

    out_stk = []
    for i in range(len(columns_cumul) - 1):
        slic = slice(columns_cumul[i], columns_cumul[i + 1])
        out = linein[slic]
        if i == 0:
            out_stk.append(out)
        else:
            try:
                out_stk.append(float(out))
            except ValueError:
                out_stk.append(np.nan)

    return out_stk

def read_rinex3_obs_legacy(rnx_in, set_index=None):
    """
    Read a RINEX Observation, version 3 or 4

    This legacy version is much slower but maybe mor relaible than the new one

    Parameters
    ----------
    rnx_in : see below
        input RINEX.
        can be the path of a RINEX file as string or as Path object,
        or directly the RINEX content as a string, bytes, StringIO object or a
        list of lines
    set_index : str or list of str, optional
        define the columns for the index.
        If None, the output DataFrame is "flat", with integer index
        ["epoch","prn"] for instance set the epoch and the prn as MultiIndex
        The default is None.

    Returns
    -------
    df_rnxobs : Pandas DataFrame / GeodeZYX's RINEX format
    """

    #### open rinex
    lines, epochs, filename = open_rinex(rnx_in)

    ### get the header
    lines_header = _header_reader(lines)

    #### get the systems and observations
    lines_sys = [l for l in lines_header if "SYS / # / OBS TYPES" in l]
    ## clean SYS / # / OBS TYPES
    lines_sys = [l[:60] for l in lines_sys]

    ## manage the 2 lines systems
    for il, l in enumerate(lines_sys):
        if l[0] == " ":
            lines_sys[il - 1] = lines_sys[il - 1] + l
            lines_sys.remove(l)

    #### store system and observables in a dictionnary
    dict_sys_nobs = dict()
    dict_sys_cols = dict()  # adds the prn, and LLI and SSI indicators
    obs_all_list = []

    for il, l in enumerate(lines_sys):
        sysobs = l.split()
        # sysobs[0] = system letter
        # sysobs[1] = observables number
        # sysobs[2:] = observables names

        dict_sys_nobs[sysobs[0]] = int(sysobs[1])
        ## adds the LLI and SSI indicators
        obs_w_lli_ssi = [(e, e + "_LLI", e + "_SSI") for e in sysobs[2:]]
        obs_all_list.extend([e for sublist in obs_w_lli_ssi for e in sublist])

        dict_sys_cols_tmp = [("prn",)] + obs_w_lli_ssi
        dict_sys_cols[sysobs[0]] = [e for sublist in dict_sys_cols_tmp for e in sublist]

        if len(sysobs[2:]) != int(sysobs[1]):
            log.warning(
                "difference between theorectical and actual obs nbr in the header"
            )

    ## the max number of observable (for the reading)
    nobs_max = max(dict_sys_nobs.values())

    df_all_stk = []
    #### reading the epochs
    for iepoc in tqdm(range(len(epochs)), desc="Reading " + filename):
        epoch = epochs[iepoc, 0]
        ## define the start/end indices of the epoch block
        iline_start = epochs[iepoc, 1] + 1
        if iepoc == len(epochs) - 1:
            iline_end = None
        else:
            iline_end = epochs[iepoc + 1, 1]

        lines_epoc = [re.sub(r"[\r\n]+", "", e) for e in lines[iline_start:iline_end]]

        # .replace("\r", "").replace("\n", "") i.e. re.sub(r"[\r\n]+", "", s) IS NOT .strip() !!
        # .strip() removes the spaces at the beginning and end of the string, and they should be kept!

        ## read the epoch block using pandas' fixed width reader
        b = StringIO("\n".join(lines_epoc))

        columns_width = [3] + nobs_max * [14, 1, 1]
        df_epoch = pd.read_fwf(b, header=None, widths=columns_width)

        # one add the system to use groupby
        #df_epoch_sysadded = df_epoch.copy()
        #df_epoch_sysadded["systmp"] = df_epoch_sysadded[0].str[0]

        df_epoch_ok_stk = []
        #### assign the correct observable names for each system
        for sys in dict_sys_cols.keys():
        #for sys, df_epoch_sys in df_epoch_sysadded.groupby("systmp"):
            df_epoch_sys = df_epoch[df_epoch[0].str[0] == sys]
            df_epoch_sys_clean = df_epoch_sys.iloc[:, : len(dict_sys_cols[sys])]
            df_epoch_sys_clean.columns = dict_sys_cols[sys]
            df_epoch_ok_stk.append(df_epoch_sys_clean)

        df_epoch_ok = pd.concat(df_epoch_ok_stk)
        # An epoch column is created to fasten the process
        col_epoch = pd.Series([epoch] * len(df_epoch_ok), name="epoch")
        col_sys = pd.Series(df_epoch_ok["prn"].str[0], name="sys")
        col_prni = pd.Series(df_epoch_ok["prn"].str[1:], name="prni", dtype=int)
        df_epoch_ok = pd.concat([col_epoch, col_sys, col_prni, df_epoch_ok], axis=1)

        df_all_stk.append(df_epoch_ok)

    ## final concat and cosmetic (reorder columns, sort)
    df_rnx_obs = pd.concat(df_all_stk)
    main_cols = ["epoch", "sys", "prn", "prni"]
    obs_used_list = list(set(df_rnx_obs.columns).difference(set(main_cols)))
    df_rnx_obs = df_rnx_obs.reindex(main_cols + list(sorted(obs_used_list)), axis=1)

    df_rnx_obs.sort_values(["epoch", "prn"], inplace=True)
    df_rnx_obs.reset_index(drop=True, inplace=True)

    if set_index:
        df_rnx_obs.set_index(set_index, inplace=True)
        df_rnx_obs.sort_index(inplace=True)

    return df_rnx_obs


if __name__ == "__main__" and False:

    # p3 = "/home/psakicki/aaa_FOURBI/RINEXexemple/ENOG00REU_R_20243030000_01D_30S_MO.rnx"
    # df3 = read_rinex3_obs(p3)
    # print(df3)

    time_readrinex = """
from __main__ import read_rinex2_obs
p = "/home/psakicki/aaa_FOURBI/RINEXexemple/smne176z.18o"
p = "/home/psakicki/aaa_FOURBI/RINEXexemple/mlvl176z.18o"
p = "/home/psakicki/Desktop/daej2220.04o"
read_rinex2_obs(p)
"""

    # import timeit
    #
    # t = timeit.timeit(time_readrinex, number=10)
    # print(t)


    #p2 = "/home/psakicki/Desktop/daej2220.04o"
    #df2 = read_rinex2_obs(p2)
    #print(df2)

    p3 = "/home/psakicki/aaa_FOURBI/RINEXexemple/ENOG00REU_R_20243030000_01D_30S_MO.rnx"
    p3 = "/home/psakicki/aaa_FOURBI/RINEXexemple/CFNG00REU_R_20240430000_15M_01S_MO.rnx"
    p3 = "/home/psakicki/aaa_FOURBI/RINEXexemple/REYK00ISL_R_20200420000_01D_30S_MO.rnx"
    df31 = read_rinex3_obs(p3)
    df32 = read_rinex3_obs_legacy(p3)
    print(df31)
    print(df32)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic
This sub-module of geodezyx.operational contains functions to download
gnss data and products from distant IGS servers. 
it can be imported directly with:
from geodezyx import operational
The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License
Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
# External modules
import datetime as dt
import glob
# Import the logger
import logging
import os
import re
import pandas as pd

# geodeZYX modules
from geodezyx import conv, utils

log = logging.getLogger('geodezyx')


##########  END IMPORT  ##########

def rinex_finder(
        main_dir,
        short_name=True,
        long_name=True,
        gfz_godc_name=True,
        compressed=None,
        specific_sites=[],
        start_epoch=None,
        end_epoch=None,
):
    """
    Find RINEX files in a specified directory and filter them based on various criteria.

    Parameters
    ----------
    main_dir : str
        Main directory where the RINEX files are stored.
        The directory can contain wildcards ('*', '?', etc.) and date aliases 
        ('%Y', '%j', etc.). If the main_dir contains date aliases and both 
        start_epoch and end_epoch are defined, you can indicate a more precise 
        directory (e.g., main_dir = "/path/to/data/*/%Y/%j/").
        Note: The day level is the maximum resolution for the wildcard.
    short_name : bool, optional
        Check if the pattern matches a short name RINEX. Default is True.
    long_name : bool, optional
        Check if the pattern matches a long name RINEX. Default is True.
    gfz_godc_name : bool, optional
        Check if the pattern matches a GFZ's GODC (GNSS Operational Data Center) 
        internal long name RINEX. Default is True.
    compressed : bool or None, optional
        Check if the pattern matches a compressed RINEX (True) or not (False).
        If None, does not matter (returns both compressed and uncompressed).
        Default is None.
    specific_sites : list of str, optional
        Filter only those specific sites. Default is [].
    start_epoch : datetime, optional
        Start date for filtering the RINEX files. Default is None.
    end_epoch : datetime, optional
        End date for filtering the RINEX files. Default is None.

    Returns
    -------
    files_rnx_lis : list of str
        List of found RINEX files.

    Notes
    -----
    This function is very similar to geodetik.rinex_lister, gins_runner.get_rinex_list, 
    and operational.multi_finder_rinex. However, this one is the most recent and 
    elaborated (July 2022) and should be used in priority.
    """

    # If main_dir contains a wildcard and both start_epoch and end_epoch are defined, search for files in the date range
    if "%" in main_dir and start_epoch and end_epoch:
        files_raw_lis = []
        for epo in conv.dt_range(start_epoch, end_epoch, day_step=1):
            main_dir_abs = os.path.abspath(epo.strftime(main_dir))
            files_raw_lis_epo, _ = utils.walk_dir(main_dir_abs)
            files_raw_lis.extend(files_raw_lis_epo)
    else:
        main_dir_abs = os.path.abspath(main_dir)
        files_raw_lis, _ = utils.walk_dir(main_dir_abs)

    files_rnx_lis = []

    # Filter files based on the provided naming patterns
    for f in files_raw_lis:
        fbase = os.path.basename(f)
        regex_match = conv.rinex_regex_search_tester(
            fbase,
            short_name=short_name,
            long_name=long_name,
            gfz_godc_name=gfz_godc_name,
            compressed=compressed,
        )
        if regex_match:
            files_rnx_lis.append(f)

    # Second filtering if specific_sites list is defined
    if len(specific_sites) > 0:
        files_rnx_lis_tmp = []
        for site in specific_sites:
            for rnx in files_rnx_lis:
                rnx_bn = os.path.basename(rnx)
                if utils.is_in_str(rnx_bn, site.upper(), site[0:4].lower(), site[0:4].upper()):
                    files_rnx_lis_tmp.append(rnx)
        files_rnx_lis = files_rnx_lis_tmp

    # Third filtering if start or end epoch are defined
    if start_epoch or end_epoch:
        if not start_epoch:
            start_epoch = dt.datetime(1980, 1, 1)
        if not end_epoch:
            end_epoch = dt.datetime(2099, 1, 1)
        files_rnx_lis_tmp = []
        dates_rnx = [conv.rinexname2dt(rnx) for rnx in files_rnx_lis]

        for rnx, date in zip(files_rnx_lis, dates_rnx):
            if start_epoch <= date <= end_epoch:
                files_rnx_lis_tmp.append(rnx)
        files_rnx_lis = files_rnx_lis_tmp

    # Sort the final list of RINEX files
    files_rnx_lis = list(sorted(files_rnx_lis))
    log.info(str(len(files_rnx_lis)) + " RINEXs found")

    return files_rnx_lis


def read_rinex_list_table(rnx_list_inp):
    """
    Generate a Table from a RINEX list

    Parameters
    ----------
    rnx_list_inp : str, list or iterable
        RINEX list. Can be a Python iterable (a list) or a path to a list file
        (a string)

    Returns
    -------
    df : DataFrame
        a RINEX Table.

    Note
    ----
    From script
    .../geodezyx_toolbox_PS_perso_scripts/IPGP_OVS/rinex_lister/rinex_list_2019_2022_mk01.py

    """

    if utils.is_iterable(rnx_list_inp):
        df = pd.DataFrame(rnx_list_inp)
    else:
        df = pd.read_csv(rnx_list_inp, header=None)

    df.columns = ["path"]
    df["name"] = df["path"].apply(os.path.basename)
    df["site"] = df["name"].str[:4]
    df["date"] = df["name"].apply(conv.rinexname2dt)
    df["sd"] = list(zip(*(df["site"], df["date"])))

    return df


# def date_filter(df_inp,strt,end):
#     return df_inp[(df_inp.date >= strt) & (df_inp.date < end)].copy()

# def clean_w_year_path(df_inp):

#     DF = df_inp.copy()
#     DF["yr_path"] = DF.path.str.extract("([0-9]{4})").astype(int)
#     BOOL = DF["yr_path"] == DF["date"].dt.year
#     DF = DF[BOOL]

#     print("before/after clean_w_year_path",len(df_inp),len(DF))

#     return DF


def find_igs_products_files(
        parent_dir,
        file_type,
        ACs,
        date_start,
        date_end=None,
        recursive_search=True,
        severe=True,
        compressed="incl",
        regex_old_naming=True,
        regex_new_naming=True,
        regex_igs_tfcc_naming=True,
        add_weekly_file=False,
        add_hourly_file=False,
):
    """
    Find all product files in a parent folder which correspond to file type(s),
    AC(s) and date(s)

    Parameters
    ----------
    parent_dir : str or list of str
        The parent directory (i.e. the archive) where files are stored
        can be a string (path of the archive) or a list of file paths
        (given by the function utils.find_recursive) in order to gain time


    file_type : str or list of str
        File type(s) researched (sp3, erp, clk ...)
        can be a list of string for several file type paths
        or a string like 'sp3' if only one file type is researched

    ACs : 'all' or str or list of str
        AC(s) researched
        can be a list of string for several ACs
        or a string like 'gfz' if only one AC is researched
        if 'all', search for all the ACs

    date_start : dt.datetime or 2-tuple list of int
        begining of the time period researched
        can be a datetime
        or a 2-tuple (wwww,d) e.g. (1990,0)

    date_end : None or dt.datetime or 2-tuple list of int
        end of the time period researched
        can be a datetime
        or a 2-tuple (wwww,d) e.g. (1990,0)
        if None, then only date_start is researched

    severe : bool
        If True, raises an exception if something goes wrong

    compressed : str
        How the compressed files are handled
        "incl": include the compressed files
        "only": only consider the compressed files
        "excl": exclude the compressed files

    regex_old_naming : bool
        Handle old naming format

    regex_new_naming : bool
        Handle new naming format

    regex_igs_tfcc_naming : bool
        Handle TFCC specific format (for SINEX files)

    add_weekly_file : bool
        Also handle the weekly file (day 7)
        Implemented only for the  old naming format (for the moment)

    add_hourly_file : bool
        Also handle ultra rapide hourly file
        Implemented only for the new naming format

    Returns
    -------
    Files_select_cumul_list : list
        List of files found

    """

    ###### Prelim Checks   ##############
    if not os.path.exists(parent_dir):
        log.error("parent directory doesn't exist")
        log.error(parent_dir)
        if severe:
            raise Exception

    ###### Time management ##############
    # work internally with datetime
    # date_start
    if type(date_start) is dt.datetime:
        date_start_ok = date_start
    else:
        date_start_ok = conv.gpstime2dt(*date_start)
    # date_end
    if not date_end:
        date_end_ok = date_start_ok
    elif type(date_end) is dt.datetime:
        date_end_ok = date_end
    else:
        date_end_ok = conv.gpstime2dt(*date_end)
    # generate time period with a while loop
    Dates_list = [date_start_ok]

    if add_hourly_file:
        deltat = dt.timedelta(seconds=3600)
    else:
        deltat = dt.timedelta(days=1)

    while Dates_list[-1] < date_end_ok:
        Dates_list.append(Dates_list[-1] + deltat)

    # manage weekly file
    dates_wwwwd_list = [
        utils.join_improved("", *conv.dt2gpstime(d, outputtype=str)) for d in Dates_list
    ]

    if add_hourly_file:
        dates_yyyyddd_list = [
            utils.join_improved(
                "", *reversed(conv.dt2doy_year(d)), str(d.hour).zfill(2)
            )
            for d in Dates_list
        ]
    else:
        dates_yyyyddd_list = [
            utils.join_improved("", *reversed(conv.dt2doy_year(d))) for d in Dates_list
        ]

    ###### File type / ACs management ##############

    if not utils.is_iterable(file_type):
        file_type = [file_type]
    if not utils.is_iterable(ACs):
        ACs = [ACs]

    ###### General file search management ##############
    if utils.is_iterable(parent_dir):
        file_list = parent_dir
    elif recursive_search:
        # All the files are listed first
        file_list = utils.find_recursive(parent_dir, ".*", regex=True)
    else:
        file_list = glob.glob(parent_dir + "/*")

    ###### Regex Definition ##############

    def join_regex_and(L):
        return "(" + "|".join(L) + ")"

    re_patt_big_stk = []

    # compression handeling
    if compressed == "excl":
        re_patt_comp = "$"
    elif compressed == "incl":
        re_patt_comp = r"(\.Z|\.gz|)$"
    elif compressed == "only":
        re_patt_comp = r"(\.Z|\.gz)$"
    else:
        log.error("check 'compressed' keyword (excl,incl, or only)")
        raise Exception

    if regex_old_naming:
        if ACs[0] == "all":
            re_patt_ac = r"\w{3}"
        else:
            re_patt_ac = join_regex_and([ac.lower() for ac in ACs])

        if add_weekly_file:
            dates_wwwwd_list_4old = [
                e[:-1] + "(" + e[-1] + "|" + "7)" for e in dates_wwwwd_list
            ]
        else:
            dates_wwwwd_list_4old = dates_wwwwd_list

        re_patt_date = join_regex_and(dates_wwwwd_list_4old)
        re_patt_filtyp = join_regex_and(file_type)
        re_patt_big_old_naming = (
                re_patt_ac + re_patt_date + r"\." + re_patt_filtyp + re_patt_comp
        )
        re_patt_big_stk.append(re_patt_big_old_naming)

    if regex_new_naming:  # search for new name convention
        if ACs[0] == "all":
            re_patt_ac = r"\W{3}"
        else:
            re_patt_ac = join_regex_and([ac.upper() for ac in ACs])
        # add _ because it can raise a conflit with the old format
        re_patt_date = join_regex_and(["_" + e for e in dates_yyyyddd_list])
        re_patt_filtyp = join_regex_and([fil.upper() for fil in file_type])
        re_patt_big_new_naming = ".*".join(
            (re_patt_ac, re_patt_date, re_patt_filtyp + re_patt_comp)
        )
        re_patt_big_stk.append(re_patt_big_new_naming)

    if regex_igs_tfcc_naming:
        dates_yy_list = list(
            set(
                [
                    str(conv.gpstime2dt(int(e[0:4]), int(e[4])).year)[2:]
                    for e in dates_wwwwd_list
                ]
            )
        )
        dates_wwww_list = list(set([e[:-1] for e in dates_wwwwd_list]))
        # Dates_wwww_dot_list = [e + "\." for e in Dates_wwww_list]
        re_patt_year = join_regex_and(dates_yy_list)
        # 2x re_patt_date : because .sum doesn't the day
        re_patt_date = join_regex_and(dates_wwwwd_list + dates_wwww_list)
        re_patt_filtyp = r"\." + join_regex_and(file_type)

        re_patt_big_igs_tfcc_naming = (
                "igs"
                + re_patt_year
                + "p"
                + re_patt_date
                + ".*"
                + re_patt_filtyp
                + re_patt_comp
        )
        re_patt_big_stk.append(re_patt_big_igs_tfcc_naming)

    re_patt_big = join_regex_and(re_patt_big_stk)

    # log.info("REGEX researched: %s",re_patt_big)

    ###### Specific file search management ##############
    files_select_list = []
    for fil in file_list:
        if re.search(re_patt_big, os.path.basename(fil), re.IGNORECASE):
            files_select_list.append(fil)

    if len(files_select_list) == 0:
        log.error("no products found")
        if severe:
            raise Exception

    log.info("%i files found matching REGEX %s", len(files_select_list), re_patt_big)

    return files_select_list

##### FUNCTION GRAVEYARD #####

# def multi_finder_rinex(
#     main_dir, rinex_types=("o", "d", "d.Z", "d.z"), specific_stats=[]
# ):
#     """
#     from a main_dir, find all the rinexs in this folder and his subfolder
#     (corresponding to the rinex_types)
#     and return a list of the found rinexs
#
#     Designed for RINEX-2 / short names only
#
#     is very similar with geodetik.rinex_lister,  gins_runner.get_rinex_list,
#     operational.rinex_finder
#
#     operational.rinex_finder must be used in priority !!! (July 2022)
#     """
#
#     log.warning("multi_finder_rinex depreciated, use rinex_finder instead!!")
#
#     files_raw_lis, _ = utils.walk_dir(main_dir)
#
#     yylis = [
#         str(e).zfill(2)
#         for e in list(range(80, 100))
#         + list(range(0, dt.datetime.now().year - 2000 + 1))
#     ]
#
#     rinex_lis = []
#
#     for f in files_raw_lis:
#         for rnxext in rinex_types:
#             for yy in yylis:
#                 if f.endswith(yy + rnxext):
#                     rinex_lis.append(f)
#
#     # CASE FOR specific_stats
#     if len(specific_stats) > 0:
#         rinex_lis2 = []
#         for stat in specific_stats:
#             for rnx in rinex_lis:
#                 if stat in os.path.basename(rnx):
#                     rinex_lis2.append(rnx)
#         rinex_lis = rinex_lis2
#
#     log.info(str(len(rinex_lis)) + " RINEXs found")
#
#     return rinex_lis



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.operational contains functions to run the 
GNSS processing software TRACK. 

it can be imported directly with:
from geodezyx import operational

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""


########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import glob
#### Import the logger
import logging
import os
import shutil
import subprocess

#### geodeZYX modules
from geodezyx import conv
from geodezyx import files_rw
from geodezyx import operational
from geodezyx import utils

log = logging.getLogger('geodezyx')

##########  END IMPORT  ##########


def track_runner(
    rnx_rover,
    rnx_base,
    working_dir,
    experience_prefix,
    XYZbase=[],
    XYZrover=[],
    outtype="XYZ",
    mode="short",
    interval=None,
    antmodfile="~/gg/tables/antmod.dat",
    calc_center="igs",
    forced_sp3_path="",
    const="G",
    silent=False,
    rinex_full_path=False,
    run_on_gfz_cluster=False,
    forced_iono_path="",
):

    # paths & files
    working_dir = utils.create_dir(working_dir)
    temp_dir = utils.create_dir(os.path.join(working_dir, "TEMP"))
    out_dir = utils.create_dir(os.path.join(working_dir, "OUTPUT"))

    if operational.check_if_compressed_rinex(rnx_rover):
        rnx_rover = operational.crz2rnx(rnx_rover, temp_dir)
    else:
        shutil.copy(rnx_rover, temp_dir)

    if operational.check_if_compressed_rinex(rnx_base):
        rnx_base = operational.crz2rnx(rnx_base, temp_dir)
    else:
        shutil.copy(rnx_base, temp_dir)

    # RINEX START & END
    rov_srt, rov_end, rov_itv = operational.rinex_start_end(rnx_rover, 1)
    bas_srt, bas_end, bas_itv = operational.rinex_start_end(rnx_base, 1)

    # RINEX NAMES
    rov_name = os.path.basename(rnx_rover)[0:4]
    bas_name = os.path.basename(rnx_base)[0:4]

    rov_name_uper = rov_name.upper()
    bas_name_uper = bas_name.upper()

    srt_str = rov_srt.strftime("%Y_%j")
    exp_full_name = "_".join((experience_prefix, rov_name, bas_name, srt_str))

    out_conf_fil = os.path.join(out_dir, exp_full_name + ".cmd")
    out_result_fil = os.path.join(out_dir, exp_full_name + ".out")

    log.info(out_conf_fil)

    confobj = open(out_conf_fil, "w+")

    # Obs Files
    confobj.write(" obs_file" + "\n")
    ### just the basename, the caracter nb is limited  (20210415)
    if not rinex_full_path:
        confobj.write(
            " ".join((" ", bas_name_uper, os.path.basename(rnx_base), "F")) + "\n"
        )
        confobj.write(
            " ".join((" ", rov_name_uper, os.path.basename(rnx_rover), "K")) + "\n"
        )
    else:
        confobj.write(" ".join((" ", bas_name_uper, rnx_base, "F")) + "\n")
        confobj.write(" ".join((" ", rov_name_uper, rnx_rover, "K")) + "\n")
    confobj.write("\n")

    date = conv.rinexname2dt(os.path.basename(rnx_rover))

    # Nav File
    if forced_sp3_path == "":
        strt_rnd = dt.datetime(*bas_srt.timetuple()[:3])
        end_rnd = dt.datetime(*bas_end.timetuple()[:3])

        orblis = operational.multi_downloader_orbs_clks(
            temp_dir, strt_rnd, end_rnd, archtype="/", calc_center=calc_center
        )

        # sp3Z = orblis[0]
        sp3 = [utils.uncompress(sp3Z) for sp3Z in orblis]
        sp3 = [e if ".sp3" in e[-5:] else e + ".sp3" for e in sp3]
    else:
        if utils.is_iterable(forced_sp3_path):
            sp3 = forced_sp3_path
        else:
            sp3 = [forced_sp3_path]
    for sp3_mono in sp3:
        confobj.write(" ".join((" ", "nav_file", sp3_mono, " sp3")) + "\n")
    confobj.write("\n")

    # Iono file

    if forced_iono_path != "":
        confobj.write(" ionex_file " + forced_iono_path + "\n")

    # Mode
    confobj.write(" mode " + mode + "\n")
    confobj.write("\n")

    # Output
    confobj.write(" pos_root " + exp_full_name + ".pos" + "\n")
    confobj.write(" res_root " + exp_full_name + ".res" + "\n")
    confobj.write(" sum_file " + exp_full_name + ".sum" + "\n")
    confobj.write("\n")

    # Outtype
    confobj.write(" out_type " + outtype + "\n")
    confobj.write("\n")

    # Interval
    if not interval:
        confobj.write(" interval " + str(rov_itv) + "\n")
    else:
        confobj.write(" interval " + str(interval) + "\n")

    confobj.write("\n")

    # Coords
    bool_site_pos = False
    if XYZbase != []:
        if not bool_site_pos:
            confobj.write(" site_pos \n")
            bool_site_pos = True
        XYZbase = [str(e) for e in XYZbase]
        confobj.write(" ".join([" ", bas_name_uper] + XYZbase + ["\n"]))

    if XYZrover != []:
        if not bool_site_pos:
            confobj.write(" site_pos \n")
            bool_site_pos = True
        XYZrover = [str(e) for e in XYZrover]
        confobj.write(" ".join([" ", rov_name_uper] + XYZrover + ["\n"]))

    if bool_site_pos:
        confobj.write("\n")

    # Offsets
    confobj.write(" ante_off \n")

    Antobj_rov, Recobj_rov, Siteobj_rov, Locobj_rov = files_rw.read_rinex_2_dataobjts(
        rnx_rover
    )

    confobj.write(
        " ".join(
            [
                " ",
                rov_name_uper,
                str(Antobj_rov.North_Ecc),
                str(Antobj_rov.East_Ecc),
                str(Antobj_rov.Up_Ecc),
                Antobj_rov.Antenna_Type,
                "\n",
            ]
        )
    )

    Antobj_bas, Recobj_bas, Siteobj_bas, Locobj_bas = files_rw.read_rinex_2_dataobjts(
        rnx_base
    )

    confobj.write(
        " ".join(
            [
                " ",
                bas_name_uper,
                str(Antobj_bas.North_Ecc),
                str(Antobj_bas.East_Ecc),
                str(Antobj_bas.Up_Ecc),
                Antobj_bas.Antenna_Type,
                "\n",
            ]
        )
    )
    confobj.write("\n")

    # Site_stats
    confobj.write(" site_stats \n")
    confobj.write(" " + bas_name_uper + " 0.1 0.1 0.1 0 0 0" + "\n")
    confobj.write(" " + rov_name_uper + " 20 20 20 0.5 0.5 0.5" + "\n")
    confobj.write("\n")

    # constellqtions
    confobj.write(" TR_GNSS " + const + "\n")

    # Misc
    # confobj.write(" USE_GPTGMF"   + '\n')
    confobj.write(" ATM_MODELC GMF 0.5" + "\n")
    confobj.write(" ANTMOD_FILE " + antmodfile + "\n")
    confobj.write(" DCB_FILE " + "~/gg/incremental_updates/tables/dcb.dat.gnss" + "\n")

    confobj.write(" atm_stats" + "\n")
    confobj.write("  all 0.1 0.00030.00023" + "\n")

    confobj.close()
    # END OF FILE WRITING

    dowstring = "".join([str(e) for e in conv.dt2gpstime(date)])
    bigcomand = " ".join(
        ("track -f", out_conf_fil, "-d", conv.dt2doy(date), "-w", dowstring)
    )

    if run_on_gfz_cluster:
        bigcomand = "cjob -c '" + bigcomand + "'"
        executable = "/bin/csh"
    else:
        executable = "/bin/bash"

    log.info("command launched :")
    log.info(bigcomand)

    # START OF PROCESSING
    if not silent:
        os.chdir(temp_dir)
        try:
            subprocess.call(
                [bigcomand], executable=executable, shell=True, timeout=60 * 20
            )
        except subprocess.TimeoutExpired:
            log.warning("command timeout expired, skip")
            pass

        outfiles = []
        outfiles = outfiles + glob.glob(os.path.join(temp_dir, exp_full_name + "*sum*"))
        outfiles = outfiles + glob.glob(os.path.join(temp_dir, exp_full_name + "*pos*"))
        outfiles = outfiles + glob.glob(os.path.join(temp_dir, exp_full_name + "*cmd*"))

        Antobj_rov, Recobj_rov, Siteobj_rov, Locobj_rov = (
            files_rw.read_rinex_2_dataobjts(rnx_rover)
        )

        [shutil.copy(e, out_dir) for e in outfiles]
        [os.remove(e) for e in outfiles]

        log.info("TRACK RUN FINISHED")
        log.info("results available in %s", out_dir)
    else:
        log.info("Silent mode ON: nothing is launched")

    return bigcomand


def run_track(temp_dir, exp_full_name, out_conf_fil, date, rnx_rover):
    dowstring = "".join([str(e) for e in conv.dt2gpstime(date)])
    bigcomand = " ".join(
        ("track -f", out_conf_fil, "-d", conv.dt2doy(date), "-w", dowstring)
    )

    log.info("INFO : command launched :")
    log.info(bigcomand)

    # START OF PROCESSING
    os.chdir(temp_dir)
    subprocess.call([bigcomand], executable="/bin/bash", shell=True)

    outfiles = []
    outfiles = outfiles + glob.glob(os.path.join(temp_dir, exp_full_name + "*sum*"))
    outfiles = outfiles + glob.glob(os.path.join(temp_dir, exp_full_name + "*pos*"))
    outfiles = outfiles + glob.glob(os.path.join(temp_dir, exp_full_name + "*cmd*"))

    Antobj_rov, Recobj_rov, Siteobj_rov, Locobj_rov = files_rw.read_rinex_2_dataobjts(
        rnx_rover
    )

    [shutil.copy(e, out_dir) for e in outfiles]
    [os.remove(e) for e in outfiles]

    log.info("TRACK RUN FINISHED")
    log.info("results available in %s ", out_dir)

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 06/06/2025 11:23:06

@author: psakic
"""

import argparse
import datetime as dt
import sys
from geodezyx import conv

from geodezyx.operational.download_rinex import download_gnss_rinex


def parse_args():
    parser = argparse.ArgumentParser(
        description="Download GNSS RINEX files from various data centers. "
                   "Supports multiple archive centers (IGS, EUREF, etc.) and handles "
                   "both RINEX2 and RINEX3 formats with parallel downloading capabilities."
    )

    parser.add_argument(
        "-l", "--stations_list",
        nargs="+",
        required=True,
        help="List of 4 or 9-character GNSS station names (e.g., ZIMM TLSE00FRA BRUS)"
    )
    parser.add_argument(
        "-c", "--datacenter",
        required=True,
        help="Data center/archive to download from. Supported values: "
             "'igs_cddis' or 'igs' (CDDIS data center), "
             "'igs_sopac' (SOPAC/UCSD/SIO data center), "
             "'igs_ign' (IGN's main server at St Mandé), "
             "'igs_ign_ensg' (IGN's secondary server at ENSG), "
             "'igs_bkg' (BKG's IGS data center), "
             "'rgp' (IGN's RGP main server at St Mandé), "
             "'rgp_ensg' (IGN's RGP, secondary server at ENSG, Marne-la-Vallée), "
             "'sonel' (SONEL data center), "
             "'euref' (EPN data center hosted at ROB), "
             "'nav' or 'brdc' (navigation files from ROB server), "
             "'nav_rt' or 'brdc_rt' (real-time navigation files from BKG server)"
    )
    parser.add_argument(
        "-o", "--output_dir",
        required=True,
        help="Root directory where RINEX files will be stored in organized subdirectories"
    )
    parser.add_argument(
        "-s", "--startdate",
        required=True,
        help="Start date for download period (YYYY-MM-DD or YYYY-DDD format)"
    )
    parser.add_argument(
        "-e", "--enddate",
        required=True,
        help="End date for download period (YYYY-MM-DD or YYYY-DDD format)"
    )
    parser.add_argument(
        "-a", "--archtype",
        default="stat",
        help="Archive directory structure type. Options: 'stat' (station-based), "
             "'year' (year-based), 'daily' (daily structure). "
             "Exemple: 'stat/year/doy'. "
             "Default: 'year/doy'"
    )
    parser.add_argument(
        "-u", "--user",
        default="anonymous",
        help="FTP username for authentication. Use 'anonymous' for public access"
    )
    parser.add_argument(
        "-p", "--passwd",
        default="anonymous@isp.com",
        help="FTP password for authentication. Use email for anonymous access"
    )
    parser.add_argument(
        "-nr2", "--no_rnx2",
        action="store_true",
        help="Skip downloading RINEX2 format files (short filename convention)"
    )
    parser.add_argument(
        "-nr3", "--no_rnx3",
        action="store_true",
        help="Skip downloading RINEX3 format files (long filename convention)"
    )
    parser.add_argument(
        "-f", "--force",
        action="store_true",
        help="Force re-download files even if they already exist locally"
    )
    parser.add_argument(
        "-q", "--quiet",
        action="store_true",
        help="Quiet mode: list available files on server without downloading them"
    )
    parser.add_argument(
        "-pd", "--parallel_download",
        type=int,
        default=1,
        help="Number of parallel download threads to use. Higher values speed up "
             "downloads but increase server load. Recommended: 1-8. Default: 1"
    )

    return parser.parse_args()

def main():
    args = parse_args()
    try:
        srtdate, enddate = conv.minmax_pattern_dt(args.startdate, args.enddate)
    except Exception as e:
        print(f"Error parsing dates: {e}")
        sys.exit(1)

    statdico = {args.datacenter: args.stations_list}

    download_gnss_rinex(
        statdico=statdico,
        output_dir=args.output_dir,
        startdate=srtdate,
        enddate=enddate,
        archtype=args.archtype,
        user=args.user,
        passwd=args.passwd,
        no_rnx2=args.no_rnx2,
        no_rnx3=args.no_rnx3,
        force=args.force,
        quiet_mode=args.quiet,
        parallel_download=args.parallel_download,
    )


if __name__ == "__main__":
    main()
